from unittest.mock import Mock

import pytest

from streamlink.plugins.youtube import YouTube
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlYouTube(PluginCanHandleUrl):
    __plugin__ = YouTube

    should_match_groups = [
        (
            ("default", "https://www.youtube.com/v/aqz-KE-bpKQ"),
            {"video_id": "aqz-KE-bpKQ"},
        ),
        (
            ("default", "https://www.youtube.com/live/aqz-KE-bpKQ"),
            {"video_id": "aqz-KE-bpKQ"},
        ),
        (
            ("default", "https://www.youtube.com/watch?foo=bar&baz=qux&v=aqz-KE-bpKQ&asdf=1234"),
            {"video_id": "aqz-KE-bpKQ"},
        ),
        (
            ("channel", "https://www.youtube.com/CHANNELNAME"),
            {"channel": "CHANNELNAME"},
        ),
        (
            ("channel", "https://www.youtube.com/CHANNELNAME/live"),
            {"channel": "CHANNELNAME", "live": "/live"},
        ),
        (
            ("channel", "https://www.youtube.com/@CHANNELNAME"),
            {"channel": "CHANNELNAME"},
        ),
        (
            ("channel", "https://www.youtube.com/@CHANNELNAME/live"),
            {"channel": "CHANNELNAME", "live": "/live"},
        ),
        (
            ("channel", "https://www.youtube.com/c/CHANNELNAME"),
            {"channel": "CHANNELNAME"},
        ),
        (
            ("channel", "https://www.youtube.com/c/CHANNELNAME/live"),
            {"channel": "CHANNELNAME", "live": "/live"},
        ),
        (
            ("channel", "https://www.youtube.com/channel/CHANNELID"),
            {"channel": "CHANNELID"},
        ),
        (
            ("channel", "https://www.youtube.com/channel/CHANNELID/live"),
            {"channel": "CHANNELID", "live": "/live"},
        ),
        (
            ("channel", "https://www.youtube.com/user/CHANNELNAME"),
            {"channel": "CHANNELNAME"},
        ),
        (
            ("channel", "https://www.youtube.com/user/CHANNELNAME/live"),
            {"channel": "CHANNELNAME", "live": "/live"},
        ),
        (
            ("embed", "https://www.youtube.com/embed/aqz-KE-bpKQ"),
            {"video_id": "aqz-KE-bpKQ"},
        ),
        (
            ("embed", "https://www.youtube.com/embed/live_stream?channel=CHANNELNAME"),
            {"live": "CHANNELNAME"},
        ),
        (
            ("shorthand", "https://youtu.be/aqz-KE-bpKQ"),
            {"video_id": "aqz-KE-bpKQ"},
        ),
    ]

    should_not_match = [
        "https://accounts.google.com/",
        "https://www.youtube.com",
        "https://www.youtube.com/",
        "https://www.youtube.com/feed/guide_builder",
        "https://www.youtube.com/t/terms",
        "https://youtu.be",
        "https://youtu.be/",
        "https://youtu.be/c/CHANNELNAME",
        "https://youtu.be/c/CHANNELNAME/live",
    ]


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        ("http://gaming.youtube.com/watch?v=0123456789A", "https://www.youtube.com/watch?v=0123456789A"),
        ("http://youtu.be/0123456789A", "https://www.youtube.com/watch?v=0123456789A"),
        ("http://youtube.com/embed/0123456789A", "https://www.youtube.com/watch?v=0123456789A"),
        ("http://youtube.com/embed/live_stream?channel=CHANNELID", "https://www.youtube.com/channel/CHANNELID/live"),
        ("http://www.youtube.com/watch?v=0123456789A", "https://www.youtube.com/watch?v=0123456789A"),
    ],
)
def test_translate_url(url, expected):
    assert YouTube(Mock(), url).url == expected



from __future__ import annotations

import base64
import hashlib
import re

# noinspection PyProtectedMember
from importlib.metadata import FileHash, PackagePath
from pathlib import Path
from unittest.mock import Mock, call

import pytest

import streamlink.plugins
import tests.plugin
from streamlink.options import Argument, Arguments

# noinspection PyProtectedMember
from streamlink.plugin.plugin import (
    HIGH_PRIORITY,
    LOW_PRIORITY,
    NO_PRIORITY,
    NORMAL_PRIORITY,
    Matcher,
    Matchers,
    Plugin,
    pluginargument,
    pluginmatcher,
)
from streamlink.session import Streamlink
from streamlink.session.plugins import StreamlinkPlugins
from streamlink.utils.args import boolean, comma_list_filter
from tests.plugin.testplugin import TestPlugin as _TestPlugin


PATH_BUILTINPLUGINS = Path(streamlink.plugins.__path__[0])
PATH_TESTPLUGINS = Path(tests.plugin.__path__[0])
PATH_TESTPLUGINS_OVERRIDE = PATH_TESTPLUGINS / "override"


class _Plugin(Plugin):
    def _get_streams(self):  # pragma: no cover
        pass


@pytest.fixture(autouse=True)
def caplog(caplog: pytest.LogCaptureFixture) -> pytest.LogCaptureFixture:
    caplog.set_level(1, "streamlink")
    return caplog


@pytest.fixture(scope="module")
def fake_plugin():
    @pluginmatcher(re.compile(r"fake"))
    @pluginargument("foo")
    @pluginargument("bar")
    class FakePlugin(_Plugin):
        __module__ = "streamlink.plugins.fake"

    return FakePlugin


def test_empty(caplog: pytest.LogCaptureFixture, session: Streamlink):
    assert session.plugins.get_names() == []
    assert session.plugins.get_loaded() == {}
    assert caplog.record_tuples == []


def test_set_get_del(session: Streamlink, fake_plugin: type[Plugin]):
    assert "fake" not in session.plugins

    session.plugins["fake"] = fake_plugin
    assert "fake" in session.plugins
    assert session.plugins["fake"] is fake_plugin
    assert session.plugins.get_names() == ["fake"]
    assert session.plugins.get_loaded() == {"fake": fake_plugin}
    assert session.plugins.get_loaded() is not session.plugins.get_loaded()

    del session.plugins["fake"]
    assert "fake" not in session.plugins
    assert session.plugins.get_names() == []
    assert session.plugins.get_loaded() == {}


def test_update_clear(session: Streamlink, fake_plugin: type[Plugin]):
    assert "fake" not in session.plugins

    session.plugins.update({"fake": fake_plugin})
    assert "fake" in session.plugins
    assert session.plugins["fake"] is fake_plugin
    assert session.plugins.get_names() == ["fake"]
    assert session.plugins.get_loaded() == {"fake": fake_plugin}

    session.plugins.clear()
    assert "fake" not in session.plugins
    assert session.plugins.get_names() == []
    assert session.plugins.get_loaded() == {}


def test_iter_arguments(session: Streamlink, fake_plugin: type[Plugin]):
    session.plugins.update({"fake": fake_plugin})
    assert [(name, [arg.argument_name(name) for arg in args]) for name, args in session.plugins.iter_arguments()] == [
        ("fake", ["--fake-foo", "--fake-bar"]),
    ]


class TestLoad:
    def test_load_builtin(self, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, fake_plugin: type[Plugin]):
        mock = Mock(return_value={"fake": fake_plugin})
        monkeypatch.setattr(StreamlinkPlugins, "_load_plugins_from_path", mock)
        session = Streamlink(plugins_builtin=True, plugins_lazy=False)

        assert mock.call_args_list == [call(PATH_BUILTINPLUGINS)]
        assert "fake" in session.plugins
        assert session.plugins.get_names() == ["fake"]
        assert session.plugins.get_loaded() == {"fake": fake_plugin}
        assert session.plugins["fake"].__module__ == "streamlink.plugins.fake"
        assert caplog.record_tuples == []

    def test_load_path_empty(self, tmp_path: Path, caplog: pytest.LogCaptureFixture, session: Streamlink):
        assert not session.plugins.load_path(tmp_path)
        assert session.plugins.get_names() == []
        assert session.plugins.get_loaded() == {}
        assert caplog.record_tuples == []

    def test_load_path_testplugins(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        assert session.plugins.load_path(PATH_TESTPLUGINS)
        assert "testplugin" in session.plugins
        assert "testplugin_invalid" not in session.plugins
        assert "testplugin_missing" not in session.plugins
        assert session.plugins.get_names() == ["testplugin"]
        assert session.plugins["testplugin"].__name__ == "TestPlugin"
        assert session.plugins["testplugin"].__module__ == "streamlink.plugins.testplugin"
        assert caplog.record_tuples == []

    def test_load_path_testplugins_override(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        assert session.plugins.load_path(PATH_TESTPLUGINS)
        assert "testplugin" in session.plugins
        assert caplog.record_tuples == []

        assert session.plugins.load_path(PATH_TESTPLUGINS_OVERRIDE)
        assert "testplugin" in session.plugins
        assert session.plugins.get_names() == ["testplugin"]
        assert session.plugins["testplugin"].__name__ == "TestPluginOverride"
        assert session.plugins["testplugin"].__module__ == "streamlink.plugins.testplugin"
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.session",
                "info",
                f"Plugin testplugin is being overridden by {PATH_TESTPLUGINS_OVERRIDE / 'testplugin.py'}"
                + " (sha256:47d9c5ec4167565db13aae76f2fbedcf961f5fca386e60c4384ee5e92714b5f3)",
            ),
        ]

    def test_load_path_testplugins_override_matchers(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        assert _TestPlugin.matchers
        session.plugins._matchers.update({"testplugin": _TestPlugin.matchers})

        assert "testplugin" not in session.plugins
        assert session.plugins.get_names() == ["testplugin"]
        assert caplog.record_tuples == []

        assert session.plugins.load_path(PATH_TESTPLUGINS)
        assert session.plugins.get_names() == ["testplugin"]
        assert session.plugins["testplugin"].__name__ == "TestPlugin"
        assert session.plugins["testplugin"].__module__ == "streamlink.plugins.testplugin"
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.session",
                "info",
                f"Plugin testplugin is being overridden by {PATH_TESTPLUGINS / 'testplugin.py'}"
                + " (sha256:088c9f6ddbe5ff046c0ea1ce0cacb7baff46189d153cf3f149b2e023ddf66f6c)",
            ),
        ]

    def test_importerror(self, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, session: Streamlink):
        monkeypatch.setattr("importlib.machinery.FileFinder.find_spec", Mock(return_value=None))
        assert not session.plugins.load_path(PATH_TESTPLUGINS)
        assert "testplugin" not in session.plugins
        assert session.plugins.get_names() == []
        assert [(record.name, record.levelname, record.message, bool(record.exc_info)) for record in caplog.records] == [
            (
                "streamlink.session",
                "error",
                f"Failed to load plugin testplugin from {PATH_TESTPLUGINS}\n",
                True,
            ),
            (
                "streamlink.session",
                "error",
                f"Failed to load plugin testplugin_invalid from {PATH_TESTPLUGINS}\n",
                True,
            ),
            (
                "streamlink.session",
                "error",
                f"Failed to load plugin testplugin_missing from {PATH_TESTPLUGINS}\n",
                True,
            ),
        ]

    def test_syntaxerror(self, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, session: Streamlink):
        monkeypatch.setattr("importlib.machinery.SourceFileLoader.exec_module", Mock(side_effect=SyntaxError))
        with pytest.raises(SyntaxError):
            session.plugins.load_path(PATH_TESTPLUGINS)
        assert session.plugins.get_names() == []
        assert caplog.record_tuples == []


class TestLoadPluginsData:
    @pytest.fixture()
    def session(self, monkeypatch: pytest.MonkeyPatch, fake_plugin: type[Plugin], metadata_files: Mock):
        class MockStreamlinkPlugins(StreamlinkPlugins):
            def load_builtin(self):
                self._plugins.update({"fake": fake_plugin})
                return True

        monkeypatch.setattr("streamlink.session.session.StreamlinkPlugins", MockStreamlinkPlugins)

        return Streamlink(plugins_builtin=True, plugins_lazy=True)

    @pytest.fixture(autouse=True)
    def metadata_files(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch, pluginsdata: str | None):
        options = getattr(request, "param", {})
        package_record = options.get("package-record", True)
        mode = options.get("package-record-hash-mode", "sha256")
        filehash = options.get("package-record-hash-value", None)

        mock = Mock(return_value=[])
        monkeypatch.setattr("importlib.metadata.files", mock)

        if not package_record or pluginsdata is None:
            yield mock
            return

        if filehash is None:
            # https://packaging.python.org/en/latest/specifications/recording-installed-packages/#the-record-file
            digest = hashlib.sha256(pluginsdata.encode("utf-8")).digest()
            filehash = base64.urlsafe_b64encode(digest).decode("utf-8").rstrip("=")

        packagepath = PackagePath("streamlink", "plugins", "_plugins.json")
        packagepath.hash = FileHash(f"{mode}={filehash}")

        mock.return_value = [packagepath]
        yield mock

        assert mock.call_args_list == [call("streamlink")]

    @pytest.fixture()
    def pluginsdata(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
        data = getattr(request, "param", "{}")

        mock = Mock()
        monkeypatch.setattr("streamlink.session.plugins._PLUGINSDATA_PATH", mock)

        if data is None:
            mock.read_bytes.side_effect = FileNotFoundError
        else:
            mock.read_bytes.return_value = data.encode("utf-8")

        return data

    # noinspection PyTestParametrized
    @pytest.mark.parametrize(
        ("metadata_files", "pluginsdata", "logs"),
        [
            pytest.param(
                {},
                None,
                [],
                id="empty-json-file",
            ),
            pytest.param(
                {"package-record": False},
                """{}""",
                [],
                id="no-package-record",
            ),
            pytest.param(
                {"package-record-hash-mode": "unknown"},
                """{}""",
                [("streamlink.session", "error", "Unknown plugins data hash mode, falling back to loading all plugins")],
                id="invalid-package-record-hash-mode",
            ),
            pytest.param(
                {"package-record-hash-value": "invalid"},
                """{}""",
                [("streamlink.session", "error", "Plugins data checksum mismatch, falling back to loading all plugins")],
                id="invalid-package-record-hash-value",
            ),
        ],
        indirect=["metadata_files", "pluginsdata"],
    )
    def test_fallback_load_builtin(self, caplog: pytest.LogCaptureFixture, session: Streamlink, logs: list):
        assert session.plugins.get_names() == ["fake"]
        assert [(record.name, record.levelname, record.message) for record in caplog.get_records(when="setup")] == logs

    # noinspection JsonStandardCompliance
    @pytest.mark.parametrize(
        "pluginsdata",
        [
            pytest.param(
                # language=json
                """
                    // foo
                    // bar
                    {
                        "testplugin": {
                            "matchers": [
                                {"pattern": "foo"}
                            ],
                            "arguments": []
                        }
                    }
                """,
                id="json-comments",
            ),
        ],
        indirect=True,
    )
    def test_strip_json_comment(self, caplog: pytest.LogCaptureFixture, session: Streamlink, pluginsdata: str):
        assert "fake" not in session.plugins
        assert "testplugin" not in session.plugins
        assert session.plugins.get_names() == ["testplugin"]
        assert [(record.name, record.levelname, record.message) for record in caplog.get_records(when="setup")] == []

    @pytest.mark.parametrize(
        "pluginsdata",
        [
            pytest.param(
                # language=json
                """
                    {
                        "testpluginA": {
                            "matchers": [
                                {"pattern": "foo"},
                                {"pattern": "bar", "flags": 64, "priority": 10, "name": "bar"}
                            ],
                            "arguments": []
                        },
                        "testpluginB": {
                            "matchers": [
                                {"pattern": "baz"}
                            ],
                            "arguments": []
                        }
                    }
                """,
                id="matchers",
            ),
        ],
        indirect=True,
    )
    def test_matchers(self, caplog: pytest.LogCaptureFixture, session: Streamlink, pluginsdata: str):
        assert "fake" not in session.plugins
        assert "testpluginA" not in session.plugins
        assert "testpluginB" not in session.plugins
        assert session.plugins.get_names() == ["testpluginA", "testpluginB"]
        assert [(record.name, record.levelname, record.message) for record in caplog.get_records(when="setup")] == []

        matchers_a = Matchers()
        matchers_a.add(Matcher(pattern=re.compile(r"foo"), priority=NORMAL_PRIORITY, name=None))
        matchers_a.add(Matcher(pattern=re.compile(r"bar", re.VERBOSE), priority=10, name="bar"))
        matchers_b = Matchers()
        matchers_b.add(Matcher(pattern=re.compile(r"baz"), priority=NORMAL_PRIORITY, name=None))
        assert list(session.plugins.iter_matchers()) == [("testpluginA", matchers_a), ("testpluginB", matchers_b)]

    @pytest.mark.parametrize(
        "pluginsdata",
        [
            pytest.param(
                # language=json
                """
                    {
                        "success": {
                            "matchers": [{"pattern": "foo"}],
                            "arguments": []
                        },
                        "fail": {
                            "matchers": [{"pattern": {"invalid": "type"}}],
                            "arguments": []
                        }
                    }
                """,
                id="matchers",
            ),
        ],
        indirect=True,
    )
    def test_matchers_failure(self, caplog: pytest.LogCaptureFixture, session: Streamlink, pluginsdata: str):
        assert "fake" in session.plugins
        assert "success" not in session.plugins
        assert "fail" not in session.plugins
        assert session.plugins.get_names() == ["fake"]
        assert [name for name, matchers in session.plugins.iter_matchers()] == ["fake"]
        assert [(record.name, record.levelname, record.message) for record in caplog.get_records(when="setup")] == [
            ("streamlink.session", "error", "Error while loading pluginmatcher data from JSON"),
        ]

    @pytest.mark.parametrize(
        "pluginsdata",
        [
            pytest.param(
                # language=json
                """
                    {
                        "empty": {
                            "matchers": [{"pattern": "foo"}]
                        },
                        "testpluginA": {
                            "matchers": [{"pattern": "bar"}],
                            "arguments": [
                                {
                                    "name": "foo",
                                    "action": "store",
                                    "nargs": 1,
                                    "default": "foo",
                                    "choices": ["foo", "bar"],
                                    "required": true,
                                    "help": "foo",
                                    "metavar": "FOO",
                                    "dest": "oof",
                                    "argument_name": "oof"
                                },
                                {
                                    "name": "bar",
                                    "const": "bar"
                                }
                            ]
                        },
                        "testpluginB": {
                            "matchers": [{"pattern": "baz"}],
                            "arguments": [
                                {
                                    "name": "invalid",
                                    "type": "type_which_does_not_exist"
                                },
                                {
                                    "name": "bool",
                                    "type": "bool"
                                },
                                {
                                    "name": "cmf",
                                    "type": "comma_list_filter",
                                    "type_args": [["1", "2", "3"]],
                                    "type_kwargs": {"unique": true}
                                }
                            ]
                        }
                    }
                """,
                id="arguments",
            ),
        ],
        indirect=True,
    )
    def test_arguments(self, caplog: pytest.LogCaptureFixture, session: Streamlink, pluginsdata: str):
        assert "fake" not in session.plugins
        assert "testpluginA" not in session.plugins
        assert "testpluginB" not in session.plugins
        assert session.plugins.get_names() == ["empty", "testpluginA", "testpluginB"]
        assert [(record.name, record.levelname, record.message) for record in caplog.get_records(when="setup")] == []

        # arguments are added in reverse order:
        # `Arguments` does this because of the reverse order of the @pluginargument decorator
        arguments_a = Arguments()
        arguments_a.add(
            Argument(
                name="bar",
                const="bar",
            ),
        )
        arguments_a.add(
            Argument(
                name="foo",
                action="store",
                nargs=1,
                default="foo",
                choices=["foo", "bar"],
                required=True,
                help="foo",
                metavar="FOO",
                dest="oof",
                argument_name="oof",
            ),
        )
        arguments_b = Arguments()
        arguments_b.add(
            Argument(
                name="cmf",
                type=comma_list_filter(["1", "2", "3"], unique=True),
            ),
        )
        arguments_b.add(
            Argument(
                name="bool",
                type=boolean,
            ),
        )
        assert list(session.plugins.iter_arguments()) == [("testpluginA", arguments_a), ("testpluginB", arguments_b)]

    @pytest.mark.parametrize(
        "pluginsdata",
        [
            pytest.param(
                # language=json
                """
                    {
                        "success": {
                            "matchers": [{"pattern": "foo"}],
                            "arguments": [{"name": "foo"}]
                        },
                        "fail": {
                            "matchers": [{"pattern": "bar"}],
                            "arguments": [{"name": {"invalid": "type"}}]
                        }
                    }
                """,
                id="arguments",
            ),
        ],
        indirect=True,
    )
    def test_arguments_failure(self, caplog: pytest.LogCaptureFixture, session: Streamlink, pluginsdata: str):
        assert "fake" in session.plugins
        assert "success" not in session.plugins
        assert "fail" not in session.plugins
        assert session.plugins.get_names() == ["fake"]
        assert [name for name, arguments in session.plugins.iter_arguments()] == ["fake"]
        assert [(record.name, record.levelname, record.message) for record in caplog.get_records(when="setup")] == [
            ("streamlink.session", "error", "Error while loading pluginargument data from JSON"),
        ]


class TestMatchURL:
    def test_priority(self, session: Streamlink):
        @pluginmatcher(priority=HIGH_PRIORITY, pattern=re.compile(r"^(high|normal|low|no)$"))
        class HighPriority(_Plugin):
            pass

        @pluginmatcher(priority=NORMAL_PRIORITY, pattern=re.compile(r"^(normal|low|no)$"))
        class NormalPriority(_Plugin):
            pass

        @pluginmatcher(priority=LOW_PRIORITY, pattern=re.compile(r"^(low|no)$"))
        class LowPriority(_Plugin):
            pass

        @pluginmatcher(priority=NO_PRIORITY, pattern=re.compile(r"^no$"))
        class NoPriority(_Plugin):
            pass

        session.plugins.update({
            "high": HighPriority,
            "normal": NormalPriority,
            "low": LowPriority,
            "no": NoPriority,
        })

        assert session.plugins.match_url("no") == ("high", HighPriority)
        assert session.plugins.match_url("low") == ("high", HighPriority)
        assert session.plugins.match_url("normal") == ("high", HighPriority)
        assert session.plugins.match_url("high") == ("high", HighPriority)

        del session.plugins["high"]

        assert session.plugins.match_url("no") == ("normal", NormalPriority)
        assert session.plugins.match_url("low") == ("normal", NormalPriority)
        assert session.plugins.match_url("normal") == ("normal", NormalPriority)
        assert session.plugins.match_url("high") is None

    def test_no_priority(self, session: Streamlink):
        @pluginmatcher(priority=NO_PRIORITY, pattern=re.compile(r"^no$"))
        class NoPriority(_Plugin):
            pass

        session.plugins.update({
            "no": NoPriority,
        })
        assert session.plugins.match_url("no") is None


class TestMatchURLLoadLazy:
    @pytest.fixture(autouse=True)
    def _pluginsdir(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("streamlink.session.plugins._PLUGINS_PATH", PATH_TESTPLUGINS)

    @pytest.fixture()
    def _loaded_plugins(self, session: Streamlink):
        # loaded built-in or custom plugin
        session.plugins.update({"testplugin": _TestPlugin})
        assert session.plugins.get_loaded() == {"testplugin": _TestPlugin}

    @pytest.fixture()
    def _loaded_matchers(self, session: Streamlink):
        matchers = _TestPlugin.matchers
        session.plugins._matchers.update({"testplugin": matchers})
        assert session.plugins.get_loaded() == {}
        assert session.plugins.get_names() == ["testplugin"]

    @pytest.mark.usefixtures("_loaded_plugins")
    def test_loaded(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        assert session.plugins.match_url("http://test.se") == ("testplugin", _TestPlugin)
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == []

    @pytest.mark.usefixtures("_loaded_matchers")
    def test_load(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        lookup = session.plugins.match_url("http://test.se")
        assert lookup is not None
        name, plugin = lookup
        assert name == "testplugin"
        # can't compare plugin classes here due to exec_module(), so just compare module name, matchers and arguments
        assert plugin.__module__ == "streamlink.plugins.testplugin"
        assert plugin.matchers == _TestPlugin.matchers
        assert plugin.arguments == _TestPlugin.arguments
        assert "testplugin" in session.plugins.get_loaded()
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            ("streamlink.session", "debug", "Loading plugin: testplugin"),
        ]

    @pytest.mark.usefixtures("_loaded_matchers")
    def test_fail_builtin(self, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, session: Streamlink):
        path = str(PATH_TESTPLUGINS / "testplugin.py")
        mock = Mock(side_effect=ImportError("", path=path))
        monkeypatch.setattr("streamlink.session.plugins.exec_module", mock)

        assert session.plugins.match_url("http://test.se") is None
        assert session.plugins.get_loaded() == {}
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            ("streamlink.session", "debug", "Loading plugin: testplugin"),
            ("streamlink.session", "error", f"Failed to load plugin testplugin from {path}\n"),
        ]



from streamlink.utils.thread import NamedThread


def test_named_thread():
    class One(NamedThread):
        pass

    class Two(NamedThread):
        pass

    assert One().name == "One-0"
    assert One().name == "One-1"
    assert Two().name == "Two-0"
    assert Two().name == "Two-1"

    assert One(name="foo").name == "One-foo-0"
    assert One(name="foo").name == "One-foo-1"
    assert One(name="bar").name == "One-bar-0"
    assert One(name="bar").name == "One-bar-1"
    assert One().name == "One-2"



from __future__ import annotations

import gettext

# noinspection PyProtectedMember
from argparse import SUPPRESS, ArgumentError, Namespace, _StoreConstAction, _VersionAction  # noqa: PLC2701
from typing import TYPE_CHECKING, Any
from unittest.mock import Mock, call

import pytest

from streamlink.exceptions import StreamlinkDeprecationWarning as SDW
from streamlink.plugin import Plugin, pluginargument
from streamlink_cli.argparser import (
    ArgumentParser,
    build_parser,
    setup_plugin_args,
    setup_plugin_options,
    setup_session_options,
)
from streamlink_cli.console import ConsoleUserInputRequester
from streamlink_cli.exceptions import StreamlinkCLIError
from streamlink_cli.main import main as streamlink_cli_main


if TYPE_CHECKING:
    from argparse import Action
    from pathlib import Path

    from streamlink.session import Streamlink


def pytest_generate_tests(metafunc: pytest.Metafunc):
    if "action" in metafunc.fixturenames:
        # noinspection PyProtectedMember
        metafunc.parametrize(
            "action",
            [
                pytest.param(
                    action,
                    id=next((opt for opt in action.option_strings if opt.startswith("--")), action.dest),
                )
                for action in build_parser()._actions
                if action.help != SUPPRESS
            ],
        )


@pytest.fixture(scope="module")
def parser():
    return build_parser()


def test_metavar_or_noargumentvalue(action: Action):
    assert (
        action.metavar  # has an explicit metavar description
        or isinstance(action, (_StoreConstAction, _VersionAction))  # doesn't expect a value
    )


class TestConfigFileArguments:
    @pytest.fixture()
    def parsed(self, request: pytest.FixtureRequest, parser: ArgumentParser, tmp_path: Path):
        content = "\n".join([
            "",
            " ",
            "# comment",
            "! comment",
            "invalid_option_format",
            *getattr(request, "param", []),
        ])

        config = tmp_path / "config"
        with config.open("w") as fd:
            fd.write(content)

        return parser.parse_args([f"@{config}"])

    @pytest.mark.parametrize("parsed", [[]], indirect=True)
    def test_nooptions(self, parsed: Namespace):
        assert parsed.ipv4 is None
        assert parsed.player_fifo is False
        assert parsed.player_args == ""
        assert parsed.title is None

    @pytest.mark.parametrize(
        "parsed",
        [
            pytest.param(["4"], id="shorthand name"),
            pytest.param(["ipv4"], id="full name"),
        ],
        indirect=True,
    )
    def test_alphanumerical(self, parsed: Namespace):
        assert parsed.ipv4 is True

    @pytest.mark.parametrize(
        "parsed",
        [
            pytest.param(["n"], id="shorthand name"),
            pytest.param(["player-fifo"], id="full name"),
        ],
        indirect=True,
    )
    def test_withoutvalue(self, parsed: Namespace):
        assert parsed.player_fifo is True

    @pytest.mark.parametrize(
        "parsed",
        [
            pytest.param(["a=foo bar "], id="shorthand name with operator"),
            pytest.param(["a = foo bar "], id="shorthand name with operator and surrounding whitespace"),
            pytest.param(["a   foo bar "], id="shorthand name without operator"),
            pytest.param(["player-args=foo bar "], id="full name with operator"),
            pytest.param(["player-args = foo bar "], id="full name with operator and surrounding whitespace"),
            pytest.param(["player-args   foo bar "], id="full name without operator"),
        ],
        indirect=True,
    )
    def test_withvalue(self, parsed: Namespace):
        assert parsed.player_args == "foo bar"

    @pytest.mark.parametrize(
        "parsed",
        [
            pytest.param(["title="], id="operator"),
            pytest.param(["title ="], id="operator with leading whitespace"),
            pytest.param(["title = "], id="operator with surrounding whitespace"),
        ],
        indirect=True,
    )
    def test_emptyvalue(self, parsed: Namespace):
        assert parsed.title == ""

    @pytest.mark.parametrize(
        "parsed",
        [
            pytest.param(["http-header=foo=bar=baz", "http-header=FOO=BAR=BAZ"], id="With operator"),
            pytest.param(["http-header foo=bar=baz", "http-header FOO=BAR=BAZ"], id="Without operator"),
        ],
        indirect=True,
    )
    def test_keyequalsvalue(self, parsed: Namespace):
        assert parsed.http_header == [("foo", "bar=baz"), ("FOO", "BAR=BAZ")]


class TestMatchArgumentOverride:
    @pytest.fixture(autouse=True)
    def _null_translations(self, monkeypatch: pytest.MonkeyPatch):
        null_translations = gettext.NullTranslations()
        monkeypatch.setattr("argparse._", null_translations.gettext)
        monkeypatch.setattr("argparse.ngettext", null_translations.ngettext)

    @pytest.fixture(scope="module")
    def parser(self):
        parser = ArgumentParser(exit_on_error=False)
        parser.add_argument("-a", "--one", dest="arg")
        parser.add_argument("-b", "--two", nargs=2)
        parser.add_argument("--one-or-more", nargs="+")

        return parser

    @pytest.mark.parametrize(
        "argv",
        [
            pytest.param(
                ["-a", "-v"],
                id="value-with-leading-dash-shorthand",
            ),
            pytest.param(
                ["--one", "-v"],
                id="value-with-leading-dash-full",
            ),
            pytest.param(
                ["--one=-v"],
                id="value-with-leading-dash-full-single-arg",
            ),
        ],
    )
    def test_match_argument(self, parser: ArgumentParser, argv: list):
        args, _ = parser.parse_known_args(argv)
        assert args.arg == "-v"

    @pytest.mark.parametrize(
        ("argv", "errormsg"),
        [
            pytest.param(
                ["--one"],
                "argument -a/--one: expected one argument",
                id="missing-value",
            ),
            pytest.param(
                ["--two"],
                "argument -b/--two: expected 2 arguments",
                id="missing-values",
            ),
            pytest.param(
                ["--one-or-more"],
                "argument --one-or-more: expected at least one argument",
                id="one-or-more",
            ),
        ],
    )
    def test_match_argument_error(self, parser: ArgumentParser, argv: list, errormsg: str):
        with pytest.raises(ArgumentError) as exc_info:
            parser.parse_known_args(argv)
        assert str(exc_info.value) == errormsg


@pytest.mark.parametrize(
    ("argv", "option", "expected"),
    [
        pytest.param(
            ["--locale", "xx_XX"],
            "locale",
            "xx_XX",
            id="Arg+value without mapper",
        ),
        pytest.param(
            ["--http-disable-dh"],
            "http-disable-dh",
            True,
            id="Arg with action=store_true",
        ),
        pytest.param(
            ["--http-no-ssl-verify"],
            "http-ssl-verify",
            False,
            id="Arg with action=store_false",
        ),
        pytest.param(
            ["--http-query-param", "foo=bar", "--http-query-param", "baz=qux"],
            "http-query-params",
            {"foo": "bar", "baz": "qux"},
            id="Arg+value with dict mapper",
        ),
        pytest.param(
            ["--http-ssl-cert-crt-key", "foo.crt", "bar.key"],
            "http-ssl-cert",
            ("foo.crt", "bar.key"),
            id="Arg+value with tuple mapper",
        ),
    ],
)
def test_setup_session_options(parser: ArgumentParser, session: Streamlink, argv: list, option: str, expected: Any):
    args = parser.parse_args(argv)
    setup_session_options(session, args)
    assert session.get_option(option) == expected


def test_setup_session_options_default_values(monkeypatch: pytest.MonkeyPatch, parser: ArgumentParser, session: Streamlink):
    mock_set_option = Mock()
    monkeypatch.setattr(session, "set_option", mock_set_option)
    args = parser.parse_args([])
    setup_session_options(session, args)
    assert session.options == session.options.defaults
    assert mock_set_option.call_args_list == [], "Value of unset session-option arg must be None and must not call set_option()"


@pytest.mark.parametrize(
    ("default", "new", "expected"),
    [
        pytest.param(False, None, False, id="Default False, unset"),
        pytest.param(True, None, True, id="Default True, unset"),
        pytest.param(False, False, False, id="Default False, set to False"),
        pytest.param(False, True, True, id="Default False, set to True"),
        pytest.param(True, False, False, id="Default True, set to False"),
        pytest.param(True, True, True, id="Default True, set to True"),
    ],
)
def test_setup_session_options_override(monkeypatch: pytest.MonkeyPatch, session: Streamlink, default, new, expected):
    arg = "NON_EXISTING_ARGPARSER_ARGUMENT"
    key = "NON-EXISTING-SESSION-OPTION-KEY"
    monkeypatch.setattr("streamlink_cli.argparser._ARGUMENT_TO_SESSIONOPTION", [(arg, key, None)])
    session.set_option(key, default)
    setup_session_options(session, Namespace(**{arg: new}))
    assert session.get_option(key) == expected


@pytest.mark.parametrize(
    ("namespace", "expected"),
    [
        pytest.param(Namespace(deprecated=None, new=123), 123, id="new-only"),
        pytest.param(Namespace(deprecated=123, new=None), 123, id="deprecated-only"),
        pytest.param(Namespace(deprecated=123, new=456), 456, id="new-overrides-deprecated"),
    ],
)
def test_setup_session_options_deprecation_override(
    monkeypatch: pytest.MonkeyPatch,
    session: Streamlink,
    namespace: Namespace,
    expected: int,
):
    arg_to_sessopt = [("deprecated", "option", None), ("new", "option", None)]
    monkeypatch.setattr("streamlink_cli.argparser._ARGUMENT_TO_SESSIONOPTION", arg_to_sessopt)
    setup_session_options(session, namespace)
    assert session.options.get_explicit("option") == expected


@pytest.mark.parametrize(
    ("argv", "option", "value", "deprecations"),
    [
        pytest.param(
            [
                "--stream-segmented-duration=123",
                "--hls-duration=321",
            ],
            "stream-segmented-duration",
            123,
            ["`hls-duration` has been deprecated in favor of the `stream-segmented-duration` option"],
            id="hls-duration",
        ),
        pytest.param(
            [
                "--stream-segmented-queue-deadline=0.0",
                "--hls-segment-queue-threshold=5.0",
            ],
            "stream-segmented-queue-deadline",
            0.0,
            ["`hls-segment-queue-threshold` has been deprecated in favor of the `stream-segmented-queue-deadline` option"],
            id="hls-segment-queue-threshold",
        ),
    ],
)
def test_setup_session_options_deprecations(
    parser: ArgumentParser,
    session: Streamlink,
    argv: list[str],
    option: str,
    value: Any,
    deprecations: list[str],
):
    """
    Test all deprecated ArgumentParser arguments that are mapped to Streamlink session options
    and test whether the order is correct, so deprecated arguments are always overridden.
    """
    args = parser.parse_args(argv)
    with pytest.warns(SDW) as warnings:
        setup_session_options(session, args)
    assert [str(dep.message) for dep in warnings.list] == deprecations
    assert session.options.get_explicit(option) == value


def test_cli_main_setup_session_options(monkeypatch: pytest.MonkeyPatch, parser: ArgumentParser, session: Streamlink):
    class StopTest(Exception):
        pass

    mock_setup_session_options = Mock()

    monkeypatch.setattr("sys.argv", [])
    monkeypatch.setattr("streamlink_cli.main.CONFIG_FILES", [])
    monkeypatch.setattr("streamlink_cli.main.logger", Mock())
    monkeypatch.setattr("streamlink_cli.main.streamlink", session)
    monkeypatch.setattr("streamlink_cli.main.build_parser", Mock(return_value=parser))
    monkeypatch.setattr("streamlink_cli.main.setup_streamlink", Mock())
    monkeypatch.setattr("streamlink_cli.main.setup_plugins", Mock())
    monkeypatch.setattr("streamlink_cli.main.log_root_warning", Mock())
    monkeypatch.setattr("streamlink_cli.main.log_current_versions", Mock())
    monkeypatch.setattr("streamlink_cli.main.log_current_arguments", Mock())
    monkeypatch.setattr("streamlink_cli.main.setup_session_options", mock_setup_session_options)
    monkeypatch.setattr("streamlink_cli.main.setup_signals", Mock(side_effect=StopTest))

    with pytest.raises(StopTest):
        streamlink_cli_main()

    assert mock_setup_session_options.call_count == 1, \
        "Has called setup_session_options() before setting up signals and running actual CLI code"  # fmt: skip
    assert mock_setup_session_options.call_args_list[0][0][0] is session
    assert isinstance(mock_setup_session_options.call_args_list[0][0][1], Namespace)


class TestSetupPluginArgsAndOptions:
    @pytest.fixture()
    def console(self):
        return Mock(
            ask=Mock(return_value="answer"),
            ask_password=Mock(return_value="password"),
        )

    @pytest.fixture()
    def parser(self):
        return ArgumentParser(add_help=False)

    @pytest.fixture()
    def plugin(self):
        # simple argument which requires namespace-name normalization
        @pluginargument("foo-bar")
        # argument with default value
        @pluginargument("baz", default=456)
        # suppressed argument
        @pluginargument("qux", help=SUPPRESS)
        # required argument with dependencies
        @pluginargument("user", required=True, requires=["pass", "captcha"])
        # sensitive argument (using console.ask_password if unset)
        @pluginargument("pass", sensitive=True)
        # argument with custom prompt (using console.ask if unset)
        @pluginargument("captcha", prompt="CAPTCHA code")
        class FakePlugin(Plugin):
            def _get_streams(self):  # pragma: no cover
                pass

        return FakePlugin

    @pytest.fixture()
    def session(self, session: Streamlink, console: Mock, parser: ArgumentParser, plugin: type[Plugin]):
        session.set_option("user-input-requester", ConsoleUserInputRequester(console))
        session.plugins["mock"] = plugin

        setup_plugin_args(session, parser)

        return session

    def test_setup_arguments(self, session: Streamlink, parser: ArgumentParser, plugin: type[Plugin]):
        group_plugins = next(
            (grp for grp in parser._action_groups if grp.title == "Plugin-specific options"),
            None,
        )  # pragma: no branch
        assert group_plugins is not None, "Adds the 'Plugin options' arguments group"
        assert group_plugins in parser.NESTED_ARGUMENT_GROUPS[None], "Adds the 'Plugin options' arguments group"

        group_plugin = next((grp for grp in parser._action_groups if grp.title == "Mock"), None)  # pragma: no branch
        assert group_plugin is not None, "Adds the 'Mock' arguments group"
        assert group_plugin in parser.NESTED_ARGUMENT_GROUPS[group_plugins], "Adds the 'Mock' arguments group"

        assert [
            item
            for action in parser._actions
            for item in action.option_strings
            if action.help != SUPPRESS
        ] == [
            "--mock-foo-bar",
            "--mock-baz",
            "--mock-user",
            "--mock-pass",
            "--mock-captcha",
        ], "Parser has all arguments registered"  # fmt: skip

    def test_setup_options_no_plugin_arguments(self, session: Streamlink, console: Mock):
        class ConcretePlugin(Plugin):
            def _get_streams(self):  # pragma: no cover
                return None

        options = setup_plugin_options(session, Namespace(), "mock", ConcretePlugin)
        assert options == {}
        assert options.defaults == {}

        assert not console.ask.called
        assert not console.ask_password.called

    def test_setup_options_no_user_input_requester(self, session: Streamlink, plugin: type[Plugin]):
        session.set_option("user-input-requester", None)
        with pytest.raises(RuntimeError) as exc_info:
            setup_plugin_options(session, Namespace(), "mock", plugin)
        assert str(exc_info.value) == "The Streamlink session is missing a UserInputRequester"

    def test_setup_options(self, recwarn: pytest.WarningsRecorder, session: Streamlink, plugin: type[Plugin], console: Mock):
        args = Namespace(
            mock_foo_bar=123,
            mock_baz=654,
            mock_qux="not-none",
            mock_user="username",
            mock_pass=None,
            mock_captcha=None,
        )
        options = setup_plugin_options(session, args, "mock", plugin)

        assert console.ask.call_args_list == [call("CAPTCHA code: ")]
        assert console.ask_password.call_args_list == [call("Enter mock pass: ")]

        assert plugin.arguments
        arg_foo = plugin.arguments.get("foo-bar")
        arg_baz = plugin.arguments.get("baz")
        arg_qux = plugin.arguments.get("qux")
        assert arg_foo
        assert arg_baz
        assert arg_qux
        assert arg_foo.default is None
        assert arg_baz.default == 456
        assert arg_qux.default is None

        assert options.get("foo-bar") == 123, "Overrides the default plugin-argument value"
        assert options.get("baz") == 654, "Uses the plugin-argument default value"
        assert options.get("qux") == "not-none", "Does not ignore values of suppressed plugin-arguments"
        assert options.get("pass") == "password"
        assert options.get("captcha") == "answer"

        options.clear()
        assert options.get("foo-bar") == arg_foo.default
        assert options.get("baz") == arg_baz.default
        assert options.get("qux") is None
        assert options.get("pass") is None
        assert options.get("captcha") is None

    def test_setup_options_deprecation_warning(self, recwarn: pytest.WarningsRecorder, session: Streamlink):
        @pluginargument("one-a", help=SUPPRESS)
        @pluginargument("one-b", default="default", help=SUPPRESS)
        @pluginargument("two-a", action="store_true", help=SUPPRESS)
        @pluginargument("two-b", action="store_true", default="default", help=SUPPRESS)
        @pluginargument("three-a", action="store_false", help=SUPPRESS)
        @pluginargument("three-b", action="store_false", default="default", help=SUPPRESS)
        @pluginargument("four-a", action="store_const", const=123, help=SUPPRESS)
        @pluginargument("four-b", action="store_const", const=123, default="default", help=SUPPRESS)
        class FakePlugin(Plugin):
            def _get_streams(self):  # pragma: no cover
                pass

        args_unset = Namespace(
            mock_one_a=None,
            mock_one_b="default",
            mock_two_a=False,
            mock_two_b="default",
            mock_three_a=True,
            mock_three_b="default",
            mock_four_a=None,
            mock_four_b="default",
        )
        setup_plugin_options(session, args_unset, "mock", FakePlugin)
        assert [(item.category, str(item.message)) for item in recwarn.list] == []

        args_set = Namespace(
            mock_one_a="value",
            mock_one_b="not-default",
            mock_two_a=True,
            mock_two_b="not-default",
            mock_three_a=False,
            mock_three_b="not-default",
            mock_four_a=123,
            mock_four_b="not-default",
        )
        options = setup_plugin_options(session, args_set, "mock", FakePlugin)
        assert dict(options.items()) == {
            "one-a": "value",
            "one-b": "not-default",
            "two-a": True,
            "two-b": "not-default",
            "three-a": False,
            "three-b": "not-default",
            "four-a": 123,
            "four-b": "not-default",
        }
        assert [(item.category, str(item.message)) for item in recwarn.list] == [
            (SDW, "The --mock-one-a plugin argument has been disabled and will be removed in the future"),
            (SDW, "The --mock-one-b plugin argument has been disabled and will be removed in the future"),
            (SDW, "The --mock-two-a plugin argument has been disabled and will be removed in the future"),
            (SDW, "The --mock-two-b plugin argument has been disabled and will be removed in the future"),
            (SDW, "The --mock-three-a plugin argument has been disabled and will be removed in the future"),
            (SDW, "The --mock-three-b plugin argument has been disabled and will be removed in the future"),
            (SDW, "The --mock-four-a plugin argument has been disabled and will be removed in the future"),
            (SDW, "The --mock-four-b plugin argument has been disabled and will be removed in the future"),
        ]

        options.clear()
        assert dict(options.items()) == {
            "one-a": None,
            "one-b": "default",
            "two-a": False,
            "two-b": "default",
            "three-a": True,
            "three-b": "default",
            "four-a": None,
            "four-b": "default",
        }

    def test_setup_options_user_input_oserror(
        self,
        session: Streamlink,
        plugin: type[Plugin],
        console: Mock,
    ):
        console.ask.side_effect = OSError("No input TTY available")
        console.ask_password.side_effect = OSError("No input TTY available")
        with pytest.raises(StreamlinkCLIError) as exc_info:
            setup_plugin_options(session, Mock(mock_user="username", mock_pass=None, mock_qux=None), "mock", plugin)
        assert str(exc_info.value) == "No input TTY available"



from __future__ import annotations

from io import BytesIO, TextIOWrapper
from textwrap import dedent
from typing import TextIO
from unittest.mock import Mock

import pytest

from streamlink_cli.console import ConsoleOutputStream
from streamlink_cli.console.console import ConsoleOutput
from streamlink_cli.console.user_input import ConsoleUserInputRequester


def getvalue(output: TextIO | None, size: int = -1):
    if output is None:
        return None

    output.seek(0)

    return output.read(size)


def build_stream(params: dict, wrap: bool = False):
    if not params.pop("exists", True):
        return None

    isatty = params.pop("isatty", True)
    params.setdefault("encoding", "utf-8")
    params.setdefault("errors", "backslashreplace")
    stream: TextIO
    orig = TextIOWrapper(BytesIO(), **params)
    if wrap:
        stream = ConsoleOutputStream(orig)
    else:
        stream = orig

    setattr(stream, "isatty", lambda: isatty)  # noqa: B010

    return stream


@pytest.fixture(autouse=True)
def stdin(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
    params = getattr(request, "param", {})
    stdin = build_stream(params)
    monkeypatch.setattr("sys.stdin", stdin)

    return stdin


@pytest.fixture(autouse=True)
def _console_output_stream(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(ConsoleOutputStream, "__new__", lambda *_, **__: object.__new__(ConsoleOutputStream))


@pytest.fixture()
def console_output(request: pytest.FixtureRequest):
    return build_stream(getattr(request, "param", {}), wrap=True)


@pytest.fixture()
def file_output(request: pytest.FixtureRequest):
    return build_stream(getattr(request, "param", {}), wrap=False)


@pytest.mark.parametrize(
    ("console_output", "file_output", "has_file_output", "expected_console_output", "expected_file_output"),
    [
        pytest.param(
            {"exists": False},
            {"exists": False},
            False,
            None,
            None,
            id="no-output-streams",
        ),
        pytest.param(
            {"isatty": True},
            {"exists": False},
            False,
            'foo\n{\n  "foo": "bar"\n}\n',
            None,
            id="console-output-is-a-tty",
        ),
        pytest.param(
            {"isatty": False},
            {"exists": False},
            False,
            'foo\n{\n  "foo": "bar"\n}\n',
            None,
            id="console-output-is-not-a-tty",
        ),
        pytest.param(
            {},
            {"isatty": True},
            False,
            'foo\n{\n  "foo": "bar"\n}\n',
            "",
            id="file-output-is-a-tty",
        ),
        pytest.param(
            {},
            {"isatty": False},
            True,
            'foo\n{\n  "foo": "bar"\n}\n',
            'foo\n{\n  "foo": "bar"\n}\n',
            id="file-output-is-not-a-tty",
        ),
    ],
    indirect=["console_output", "file_output"],
)
def test_streams(
    console_output: ConsoleOutputStream | None,
    file_output: TextIO | None,
    has_file_output: bool,
    expected_console_output: str,
    expected_file_output: str,
):
    console = ConsoleOutput()
    assert console.console_output is None
    assert console.file_output is None

    console.console_output = console_output
    console.file_output = file_output
    assert console.console_output is console_output
    assert console.file_output is (file_output if has_file_output else None)

    console.msg("foo")
    console.json = True
    console.msg_json(foo="bar")
    assert getvalue(console_output) == expected_console_output
    assert getvalue(file_output) == expected_file_output


@pytest.mark.parametrize("method_raises", ["write", "flush"])
def test_broken_stream(monkeypatch: pytest.MonkeyPatch, console_output: ConsoleOutputStream, method_raises: str):
    monkeypatch.setattr(console_output, method_raises, Mock(side_effect=BrokenPipeError))
    console = ConsoleOutput(console_output=console_output)
    console.msg("foo")
    console.msg("bar")


def test_close(console_output: ConsoleOutputStream, file_output: TextIO):
    console = ConsoleOutput(console_output=console_output, file_output=file_output)
    console.msg("foo")
    assert getvalue(console_output) == "foo\n"
    assert getvalue(file_output) == "foo\n"
    assert not console_output.closed
    assert not file_output.closed

    console.close()
    assert console_output.closed
    assert file_output.closed


class TestMessages:
    @pytest.mark.parametrize(
        ("console_output", "expected_console_output", "expected_file_output"),
        [
            pytest.param(
                {"encoding": "utf-8"},
                "Bär: 🐻\n",
                "Bär: 🐻\n",
                id="utf-8 encoding",
            ),
            pytest.param(
                {"encoding": "ascii"},
                "B\\xe4r: \\U0001f43b\n",  # Unicode character: "Bear Face" (U+1F43B)
                "Bär: 🐻\n",
                id="ascii encoding",
            ),
        ],
        indirect=["console_output"],
    )
    def test_no_json(
        self,
        console_output: ConsoleOutputStream,
        file_output: TextIO,
        expected_console_output: str,
        expected_file_output: str,
    ):
        console = ConsoleOutput(console_output=console_output, file_output=file_output)
        console.msg("Bär: 🐻")
        console.msg_json({"test": 1})
        assert getvalue(console_output) == expected_console_output
        assert getvalue(file_output) == expected_file_output

    @pytest.mark.parametrize(
        ("console_output", "expected_console_output", "expected_file_output"),
        [
            pytest.param(
                {"encoding": "utf-8"},
                '{\n  "test": "Bär: 🐻"\n}\n',
                '{\n  "test": "Bär: 🐻"\n}\n',
                id="utf-8 encoding",
            ),
            pytest.param(
                {"encoding": "ascii"},
                '{\n  "test": "B\\u00e4r: \\ud83d\\udc3b"\n}\n',  # Unicode character: "Bear Face" (U+1F43B)
                '{\n  "test": "Bär: 🐻"\n}\n',
                id="ascii encoding",
            ),
        ],
        indirect=["console_output"],
    )
    def test_json(
        self,
        console_output: ConsoleOutputStream,
        file_output: TextIO,
        expected_console_output: str,
        expected_file_output: str,
    ):
        console = ConsoleOutput(console_output=console_output, file_output=file_output, json=True)
        console.msg("foo")
        console.msg_json({"test": "Bär: 🐻"})
        assert getvalue(console_output) == expected_console_output
        assert getvalue(file_output) == expected_file_output

    def test_msg_json_object(self, console_output: ConsoleOutputStream):
        console = ConsoleOutput(console_output=console_output, json=True)
        console.msg_json(Mock(__json__=lambda: {"test": "Hello world, Γειά σου Κόσμε, こんにちは世界"}))  # noqa: RUF001
        assert getvalue(console_output) == '{\n  "test": "Hello world, Γειά σου Κόσμε, こんにちは世界"\n}\n'  # noqa: RUF001

    def test_msg_json_list(self, console_output: ConsoleOutputStream):
        console = ConsoleOutput(console_output=console_output, json=True)
        test_list = ["Hello world, Γειά σου Κόσμε, こんにちは世界", '"🐻"']  # noqa: RUF001
        console.msg_json(test_list)
        assert getvalue(console_output) == '[\n  "Hello world, Γειά σου Κόσμε, こんにちは世界",\n  "\\"🐻\\""\n]\n'  # noqa: RUF001

    def test_msg_json_merge_object(self, console_output: ConsoleOutputStream):
        console = ConsoleOutput(console_output=console_output, json=True)
        test_obj1 = {"test": 1, "foo": "foo"}
        test_obj2 = Mock(__json__=Mock(return_value={"test": 2}))
        console.msg_json(test_obj1, test_obj2, ["qux"], foo="bar", baz="qux")
        assert (
            getvalue(console_output)
            == dedent("""
                {
                  "test": 2,
                  "foo": "bar",
                  "baz": "qux"
                }
            """).lstrip()
        )
        assert list(test_obj1.items()) == [("test", 1), ("foo", "foo")]

    def test_msg_json_merge_list(self, console_output: ConsoleOutputStream):
        console = ConsoleOutput(console_output=console_output, json=True)
        test_list1 = ["foo", "bar"]
        test_list2 = Mock(__json__=Mock(return_value={"foo": "bar"}))
        console.msg_json(test_list1, ["baz"], test_list2, {"foo": "bar"}, foo="bar", baz="qux")
        assert (
            getvalue(console_output)
            == dedent("""
                [
                  "foo",
                  "bar",
                  "baz",
                  {
                    "foo": "bar"
                  },
                  {
                    "foo": "bar"
                  },
                  {
                    "foo": "bar",
                    "baz": "qux"
                  }
                ]
            """).lstrip()
        )
        assert test_list1 == ["foo", "bar"]

    @pytest.mark.parametrize(
        ("json", "supports_status_messages", "expected"),
        [
            pytest.param(True, False, [], id="json-no-status-messages"),
            pytest.param(True, True, [], id="json-status-messages"),
            pytest.param(False, False, [], id="no-json-no-status-messages"),
            pytest.param(False, True, ["foo"], id="no-json-status-messages"),
        ],
    )
    def test_msg_status(
        self,
        monkeypatch: pytest.MonkeyPatch,
        console_output: ConsoleOutputStream,
        json: bool,
        supports_status_messages: bool,
        expected: list[str],
    ):
        writes: list[str] = []
        monkeypatch.setattr(console_output, "supports_status_messages", Mock(return_value=supports_status_messages))
        monkeypatch.setattr(console_output, "write", writes.append)

        console = ConsoleOutput(console_output=console_output, json=json)
        assert console.supports_status_messages() == supports_status_messages
        console.msg_status("foo")
        assert writes == expected


class TestPrompts:
    def test_prompt_exception(self, console_output: ConsoleOutputStream):
        console = ConsoleOutput(console_output=console_output)
        with pytest.raises(BaseException) as exc_info:  # noqa: PT011
            with console._prompt():
                raise EOFError
        assert isinstance(exc_info.value, OSError)
        assert isinstance(exc_info.value.__cause__, EOFError)

    @pytest.mark.parametrize("exception", [OSError, KeyboardInterrupt])
    def test_prompt_exception_passthrough(self, console_output: ConsoleOutputStream, exception: type[Exception]):
        console = ConsoleOutput(console_output=console_output)
        with pytest.raises(BaseException) as exc_info:  # noqa: PT011
            with console._prompt():
                raise exception
        assert isinstance(exc_info.value, exception)

    def test_ask(self, monkeypatch: pytest.MonkeyPatch, console_output: ConsoleOutputStream, file_output: TextIO):
        monkeypatch.setattr("builtins.input", Mock(return_value="hello"))

        console = ConsoleOutput(console_output=console_output, file_output=file_output)
        user_input = ConsoleUserInputRequester(console)
        assert user_input.ask("test") == "hello"
        assert getvalue(console_output) == "test: "
        assert getvalue(file_output) == ""

    @pytest.mark.parametrize(
        ("stdin", "console_output", "expected"),
        [
            pytest.param({"exists": False}, {}, "^No input TTY available$", id="no-stdin"),
            pytest.param({"isatty": False}, {}, "^No input TTY available$", id="stdin-no-tty"),
            pytest.param({}, {"exists": False}, "^No output TTY available$", id="no-output"),
            pytest.param({}, {"isatty": False}, "^No output TTY available$", id="output-no-tty"),
        ],
        indirect=["stdin", "console_output"],
    )
    def test_ask_failure(
        self,
        monkeypatch: pytest.MonkeyPatch,
        stdin: TextIO,
        console_output: ConsoleOutputStream,
        expected: str,
    ):
        mock_input = Mock()
        monkeypatch.setattr("builtins.input", mock_input)

        console = ConsoleOutput(console_output=console_output)
        user_input = ConsoleUserInputRequester(console)
        with pytest.raises(OSError, match=expected):
            user_input.ask("test")
        assert mock_input.call_args_list == []

    def test_ask_password(self, monkeypatch: pytest.MonkeyPatch, console_output: ConsoleOutputStream):
        def getpass(prompt, stream):
            stream.write(prompt)
            stream.flush()

            return "hello"

        monkeypatch.setattr("streamlink_cli.console.console.getpass", getpass)

        console = ConsoleOutput(console_output=console_output)
        user_input = ConsoleUserInputRequester(console)
        assert user_input.ask_password("test") == "hello"
        assert getvalue(console_output) == "test: "

    @pytest.mark.parametrize(
        ("stdin", "console_output", "expected"),
        [
            pytest.param({"exists": False}, {}, "^No input TTY available$", id="no-stdin"),
            pytest.param({"isatty": False}, {}, "^No input TTY available$", id="stdin-no-tty"),
            pytest.param({}, {"exists": False}, "^No output TTY available$", id="no-output"),
            pytest.param({}, {"isatty": False}, "^No output TTY available$", id="output-no-tty"),
        ],
        indirect=["stdin", "console_output"],
    )
    def test_ask_password_failure(
        self,
        monkeypatch: pytest.MonkeyPatch,
        stdin: TextIO,
        console_output: ConsoleOutputStream,
        expected: str,
    ):
        mock_getpass = Mock()
        monkeypatch.setattr("streamlink_cli.console.console.getpass", mock_getpass)

        console = ConsoleOutput(console_output=console_output)
        user_input = ConsoleUserInputRequester(console)
        with pytest.raises(OSError, match=expected):
            user_input.ask_password("test")
        assert mock_getpass.call_args_list == []



from streamlink.plugins.invintus import InvintusMedia
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlInvintusMedia(PluginCanHandleUrl):
    __plugin__ = InvintusMedia

    should_match = [
        "https://player.invintus.com/?clientID=9375922947&eventID=2020031185",
        "https://player.invintus.com/?clientID=9375922947&eventID=2020031184",
        "https://player.invintus.com/?clientID=9375922947&eventID=2020031183",
        "https://player.invintus.com/?clientID=9375922947&eventID=2020031182",
        "https://player.invintus.com/?clientID=9375922947&eventID=2020031181",
    ]



from __future__ import annotations

import re
from inspect import currentframe, getframeinfo
from socket import AF_INET, AF_INET6
from unittest.mock import Mock, call

import pytest

from streamlink.exceptions import StreamlinkDeprecationWarning
from streamlink.session import Streamlink
from streamlink.session.options import StreamlinkOptions


class TestOptionsDocumentation:
    @pytest.fixture()
    def docstring(self, session: Streamlink):
        docstring = session.options.__doc__
        assert docstring is not None
        return docstring

    def test_default_option_is_documented(self, session: Streamlink, docstring: str):
        assert session.options.keys()
        for option in session.options:
            assert f"* - {option}" in docstring, f"Option '{option}' is documented"

    def test_documented_option_exists(self, session: Streamlink, docstring: str):
        options = session.options
        setters = options._MAP_SETTERS.keys()
        documented = re.compile(r"\* - (\S+)").findall(docstring)[1:]
        assert documented
        for option in documented:
            assert option in options or option in setters, f"Documented option '{option}' exists"


def test_default_objects():
    one = Streamlink(plugins_builtin=False)
    two = Streamlink(plugins_builtin=False)
    ids_one = {key: id(value) for key, value in one.options.defaults.items()}
    ids_two = {key: id(value) for key, value in two.options.defaults.items()}
    assert ids_one != ids_two


def test_session_wrapper_methods(session: Streamlink):
    session.set_option("test_option", "option")
    assert session.get_option("test_option") == "option"
    assert session.get_option("non_existing") is None


def test_session_option_set_deprecated(recwarn: pytest.WarningsRecorder, session: Streamlink):
    def get_lineno():
        frame = currentframe()
        assert frame
        assert frame.f_back
        return getframeinfo(frame.f_back).lineno

    class FakeStreamlinkOptions(StreamlinkOptions):
        _MAP_SETTERS = {
            "deprecated": StreamlinkOptions._factory_set_deprecated("new", int),
        }

    session.options = FakeStreamlinkOptions(session)
    assert session.get_option("new") is None
    assert recwarn.list == []

    session.set_option("deprecated", 123)
    lineno = get_lineno() - 1

    assert session.get_option("new") == 123
    assert [(item.filename, item.lineno, item.category, str(item.message)) for item in recwarn.list] == [
        (__file__, lineno, StreamlinkDeprecationWarning, "`deprecated` has been deprecated in favor of the `new` option"),
    ]


def test_options_locale(monkeypatch: pytest.MonkeyPatch, session: Streamlink):
    monkeypatch.setattr("locale.getlocale", lambda: ("C", None))
    assert session.get_option("locale") is None

    localization = session.localization
    assert localization.explicit is False
    assert localization.language_code == "en_US"
    assert localization.country.alpha2 == "US"
    assert localization.country.name == "United States"
    assert localization.language.alpha2 == "en"
    assert localization.language.name == "English"

    session.set_option("locale", "de_DE")
    assert session.get_option("locale") == "de_DE"

    localization = session.localization
    assert localization.explicit is True
    assert localization.language_code == "de_DE"
    assert localization.country.alpha2 == "DE"
    assert localization.country.name == "Germany"
    assert localization.language.alpha2 == "de"
    assert localization.language.name == "German"


def test_options_interface(monkeypatch: pytest.MonkeyPatch, session: Streamlink):
    mock = Mock()
    monkeypatch.setattr(session.http, "set_interface", mock)

    assert session.get_option("interface") is None

    session.set_option("interface", "my-interface")
    assert mock.call_args_list.pop() == call(interface="my-interface")
    assert session.get_option("interface") == "my-interface"

    session.set_option("interface", None)
    assert mock.call_args_list.pop() == call(interface=None)
    assert session.get_option("interface") is None

    # doesn't raise
    session.set_option("interface", None)
    assert mock.call_args_list.pop() == call(interface=None)
    assert session.get_option("interface") is None


def test_options_ipv4_ipv6(monkeypatch: pytest.MonkeyPatch, session: Streamlink):
    mock = Mock()
    monkeypatch.setattr(session.http, "set_address_family", mock)

    assert session.get_option("ipv4") is False
    assert session.get_option("ipv6") is False

    session.set_option("ipv4", True)
    assert mock.call_args_list.pop() == call(family=AF_INET)
    assert session.get_option("ipv4") is True
    assert session.get_option("ipv6") is False

    session.set_option("ipv4", False)
    assert mock.call_args_list.pop() == call(family=None)
    assert session.get_option("ipv4") is False
    assert session.get_option("ipv6") is False

    session.set_option("ipv6", True)
    assert mock.call_args_list.pop() == call(family=AF_INET6)
    assert session.get_option("ipv4") is False
    assert session.get_option("ipv6") is True

    session.set_option("ipv6", False)
    assert mock.call_args_list.pop() == call(family=None)
    assert session.get_option("ipv4") is False
    assert session.get_option("ipv6") is False

    session.set_option("ipv4", True)
    assert mock.call_args_list.pop() == call(family=AF_INET)
    session.set_option("ipv6", False)
    assert mock.call_args_list == []
    assert session.get_option("ipv4") is True
    assert session.get_option("ipv6") is False
    session.set_option("ipv4", False)
    assert mock.call_args_list.pop() == call(family=None)
    assert session.get_option("ipv4") is False
    assert session.get_option("ipv6") is False

    session.set_option("ipv6", True)
    assert mock.call_args_list.pop() == call(family=AF_INET6)
    session.set_option("ipv4", False)
    assert mock.call_args_list == []
    assert session.get_option("ipv4") is False
    assert session.get_option("ipv6") is True
    session.set_option("ipv6", False)
    assert mock.call_args_list.pop() == call(family=None)
    assert session.get_option("ipv4") is False
    assert session.get_option("ipv6") is False


def test_options_http_disable_dh(monkeypatch: pytest.MonkeyPatch, session: Streamlink):
    mock = Mock()
    monkeypatch.setattr(session.http, "disable_dh", mock)

    assert not session.get_option("http-disable-dh")

    session.set_option("http-disable-dh", True)
    assert mock.call_args_list.pop() == call(disable=True)
    assert session.get_option("http-disable-dh")

    session.set_option("http-disable-dh", False)
    assert mock.call_args_list.pop() == call(disable=False)
    assert not session.get_option("http-disable-dh")


class TestOptionsHttpProxy:
    @pytest.fixture()
    def _no_deprecation(self, recwarn: pytest.WarningsRecorder):
        yield
        assert recwarn.list == []

    @pytest.fixture()
    def _logs_deprecation(self, recwarn: pytest.WarningsRecorder):
        yield
        assert [(record.category, str(record.message), record.filename) for record in recwarn.list] == [
            (
                StreamlinkDeprecationWarning,
                "The `https-proxy` option has been deprecated in favor of a single `http-proxy` option",
                __file__,
            ),
        ]

    @pytest.mark.usefixtures("_no_deprecation")
    def test_https_proxy_default(self, session: Streamlink):
        session.set_option("http-proxy", "http://testproxy.com")

        assert session.http.proxies["http"] == "http://testproxy.com"
        assert session.http.proxies["https"] == "http://testproxy.com"

    @pytest.mark.usefixtures("_logs_deprecation")
    def test_https_proxy_set_first(self, session: Streamlink):
        session.set_option("https-proxy", "https://testhttpsproxy.com")
        session.set_option("http-proxy", "http://testproxy.com")

        assert session.http.proxies["http"] == "http://testproxy.com"
        assert session.http.proxies["https"] == "http://testproxy.com"

    @pytest.mark.usefixtures("_logs_deprecation")
    def test_https_proxy_default_override(self, session: Streamlink):
        session.set_option("http-proxy", "http://testproxy.com")
        session.set_option("https-proxy", "https://testhttpsproxy.com")

        assert session.http.proxies["http"] == "https://testhttpsproxy.com"
        assert session.http.proxies["https"] == "https://testhttpsproxy.com"

    @pytest.mark.usefixtures("_logs_deprecation")
    def test_https_proxy_set_only(self, session: Streamlink):
        session.set_option("https-proxy", "https://testhttpsproxy.com")

        assert session.http.proxies["http"] == "https://testhttpsproxy.com"
        assert session.http.proxies["https"] == "https://testhttpsproxy.com"

    @pytest.mark.usefixtures("_no_deprecation")
    def test_http_proxy_socks(self, session: Streamlink):
        session.set_option("http-proxy", "socks5://localhost:1234")

        assert session.http.proxies["http"] == "socks5://localhost:1234"
        assert session.http.proxies["https"] == "socks5://localhost:1234"

    @pytest.mark.usefixtures("_logs_deprecation")
    def test_https_proxy_socks(self, session: Streamlink):
        session.set_option("https-proxy", "socks5://localhost:1234")

        assert session.http.proxies["http"] == "socks5://localhost:1234"
        assert session.http.proxies["https"] == "socks5://localhost:1234"

    @pytest.mark.usefixtures("_no_deprecation")
    def test_get_http_proxy(self, session: Streamlink):
        session.http.proxies["http"] = "http://testproxy1.com"
        session.http.proxies["https"] = "http://testproxy2.com"
        assert session.get_option("http-proxy") == "http://testproxy1.com"

    @pytest.mark.usefixtures("_logs_deprecation")
    def test_get_https_proxy(self, session: Streamlink):
        session.http.proxies["http"] = "http://testproxy1.com"
        session.http.proxies["https"] = "http://testproxy2.com"
        assert session.get_option("https-proxy") == "http://testproxy2.com"

    @pytest.mark.usefixtures("_logs_deprecation")
    def test_https_proxy_get_directly(self, session: Streamlink):
        # The DeprecationWarning's origin must point to this call, even without the set_option() wrapper
        session.options.get("https-proxy")

    @pytest.mark.usefixtures("_logs_deprecation")
    def test_https_proxy_set_directly(self, session: Streamlink):
        # The DeprecationWarning's origin must point to this call, even without the set_option() wrapper
        session.options.set("https-proxy", "https://foo")


def test_options_http_cookies_files(monkeypatch: pytest.MonkeyPatch, session: Streamlink):
    mock = Mock()
    monkeypatch.setattr(session.http, "set_cookies_from_file", mock)

    session.set_option("http-cookies-files", ["foo", "bar"])
    assert mock.call_args_list == [call("foo"), call("bar")]


class TestOptionsKeyEqualsValue:
    @pytest.fixture()
    def option(self, request, session: Streamlink):
        key, attr = request.param
        httpsessionattr = getattr(session.http, attr)
        assert session.get_option(key) is httpsessionattr
        assert "foo" not in httpsessionattr
        assert "bar" not in httpsessionattr
        yield key
        assert httpsessionattr.get("foo") == "foo=bar"
        assert httpsessionattr.get("bar") == "123"

    @pytest.mark.parametrize(
        "option",
        [
            pytest.param(("http-cookies", "cookies"), id="http-cookies"),
            pytest.param(("http-headers", "headers"), id="http-headers"),
            pytest.param(("http-query-params", "params"), id="http-query-params"),
        ],
        indirect=["option"],
    )
    def test_dict(self, session: Streamlink, option: str):
        session.set_option(option, {"foo": "foo=bar", "bar": "123"})

    @pytest.mark.parametrize(
        ("option", "value"),
        [
            pytest.param(("http-cookies", "cookies"), "foo=foo=bar;bar=123;baz", id="http-cookies"),
            pytest.param(("http-headers", "headers"), "foo=foo=bar;bar=123;baz", id="http-headers"),
            pytest.param(("http-query-params", "params"), "foo=foo=bar&bar=123&baz", id="http-query-params"),
        ],
        indirect=["option"],
    )
    def test_string(self, session: Streamlink, option: str, value: str):
        session.set_option(option, value)


@pytest.mark.parametrize(
    ("option", "attr", "default", "value"),
    [
        ("http-ssl-cert", "cert", None, "foo"),
        ("http-ssl-verify", "verify", True, False),
        ("http-trust-env", "trust_env", True, False),
        ("http-timeout", "timeout", 20.0, 30.0),
    ],
)
def test_options_http_other(session: Streamlink, option: str, attr: str, default, value):
    httpsessionattr = getattr(session.http, attr)
    assert httpsessionattr == default
    assert session.get_option(option) == httpsessionattr

    session.set_option(option, value)
    assert session.get_option(option) == value



import unittest
from unittest.mock import MagicMock, call, patch

import pytest

from streamlink.plugins.kick import Kick, KickHLSStream, KickHLSStreamReader, KickHLSStreamWriter
from tests.mixins.stream_hls import EventedHLSStreamWriter, Playlist, Segment, TestMixinStreamHLS
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlKick(PluginCanHandleUrl):
    __plugin__ = Kick

    should_match_groups = [
        (
            ("live", "https://kick.com/LIVE_CHANNEL"),
            {"channel": "LIVE_CHANNEL"},
        ),
        (
            ("vod", "https://kick.com/video/VIDEO_ID"),
            {"vod": "VIDEO_ID"},
        ),
        (
            ("vod", "https://kick.com/VIDEO_CHANNEL/videos/VIDEO_ID"),
            {"vod": "VIDEO_ID"},
        ),
        (
            ("clip", "https://kick.com/CLIP_CHANNEL?clip=CLIP_ID&foo=bar"),
            {"channel": "CLIP_CHANNEL", "clip": "CLIP_ID"},
        ),
        (
            ("clip", "https://kick.com/CLIP_CHANNEL/clips/CLIP_ID?foo=bar"),
            {"channel": "CLIP_CHANNEL", "clip": "CLIP_ID"},
        ),
        (
            ("clip", "https://kick.com/CLIP_CHANNEL/clips/CLIP_ID?clip=CLIP_ID_FROM_QUERY_STRING&foo=bar"),
            {"channel": "CLIP_CHANNEL", "clip": "CLIP_ID"},
        ),
    ]


class SegmentPrefetch(Segment):
    def build(self, namespace):
        return f"#EXT-X-PREFETCH:{self.url(namespace)}"


class _KickHLSStreamWriter(EventedHLSStreamWriter, KickHLSStreamWriter):
    pass


class _KickHLSStreamReader(KickHLSStreamReader):
    __writer__ = _KickHLSStreamWriter


class _KickHLSStream(KickHLSStream):
    __reader__ = _KickHLSStreamReader


@patch("streamlink.stream.hls.HLSStreamWorker.wait", MagicMock(return_value=True))
class TestKickHLSStream(TestMixinStreamHLS, unittest.TestCase):
    __stream__ = _KickHLSStream
    stream: KickHLSStream

    def get_session(self, *args, **kwargs):
        session = super().get_session(*args, **kwargs)
        session.set_option("hls-live-edge", 4)

        return session

    @patch("streamlink.plugins.kick.log")
    def test_hls_low_latency_has_prefetch(self, mock_log):
        segments = self.subject(
            [
                Playlist(0, [Segment(0), Segment(1), Segment(2), Segment(3), SegmentPrefetch(4), SegmentPrefetch(5)]),
                Playlist(4, [Segment(4), Segment(5), Segment(6), Segment(7), SegmentPrefetch(8), SegmentPrefetch(9)], end=True),
            ],
            streamoptions={"low_latency": True},
        )

        assert self.session.options.get("hls-live-edge") == 2
        assert self.session.options.get("hls-segment-stream-data")

        self.await_write(6)
        data = self.await_read(read_all=True)
        assert data == self.content(segments, cond=lambda s: s.num >= 4), "Skips first four segments due to reduced live-edge"
        assert not any(self.called(s) for s in segments.values() if s.num < 4), "Doesn't download old segments"
        assert all(self.called(s) for s in segments.values() if s.num >= 4), "Downloads all remaining segments"
        assert mock_log.info.mock_calls == [
            call("Low latency streaming (HLS live edge: 2)"),
        ]

    @patch("streamlink.plugins.kick.log")
    def test_hls_no_low_latency_has_prefetch(self, mock_log):
        segments = self.subject(
            [
                Playlist(0, [Segment(0), Segment(1), Segment(2), Segment(3), SegmentPrefetch(4), SegmentPrefetch(5)]),
                Playlist(4, [Segment(4), Segment(5), Segment(6), Segment(7), SegmentPrefetch(8), SegmentPrefetch(9)], end=True),
            ],
            streamoptions={"low_latency": False},
        )

        assert self.session.options.get("hls-live-edge") == 4
        assert not self.session.options.get("hls-segment-stream-data")

        self.await_write(8)
        data = self.await_read(read_all=True)
        assert data == self.content(segments, cond=lambda s: s.num < 8), "Ignores prefetch segments"
        assert all(self.called(s) for s in segments.values() if s.num <= 7), "Ignores prefetch segments"
        assert not any(self.called(s) for s in segments.values() if s.num > 7), "Ignores prefetch segments"
        assert mock_log.info.mock_calls == []
        assert self.thread.reader.worker._reload_time == pytest.approx(3.0)

    @patch("streamlink.plugins.kick.log")
    def test_hls_low_latency_no_prefetch(self, mock_log):
        self.subject(
            [
                Playlist(0, [Segment(0), Segment(1), Segment(2), Segment(3)]),
                Playlist(4, [Segment(4), Segment(5), Segment(6), Segment(7)], end=True),
            ],
            streamoptions={"low_latency": True},
        )

        assert self.stream.low_latency

        self.await_write(6)
        self.await_read(read_all=True)
        assert mock_log.info.mock_calls == [
            call("Low latency streaming (HLS live edge: 2)"),
        ]

    def test_hls_low_latency_no_ads_reload_time(self):
        Seg, SegPre = Segment, SegmentPrefetch
        self.subject(
            [
                Playlist(0, [Seg(0, duration=5), Seg(1, duration=7), Seg(2, duration=11), SegPre(3)], end=True),
            ],
            streamoptions={"low_latency": True},
        )

        self.await_write(4)
        self.await_read(read_all=True)
        assert self.thread.reader.worker._reload_time == pytest.approx(23 / 3)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys
from pathlib import Path

from streamlink import __version__ as streamlink_version


sys.path.insert(0, str(Path(__file__).resolve().parent / "sphinxext"))


# -- Project information -------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Streamlink"
project_copyright = "2026, Streamlink"
author = "Streamlink"
version = streamlink_version.split("+")[0]
release = streamlink_version


# -- General configuration -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

needs_sphinx = "6.0"

extensions = [
    # Sphinx-internal
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    # third-party
    "myst_parser",
    "sphinx_design",
    # custom
    "ext_html_template_vars",
    "ext_argparse",
    "ext_github",
    "ext_plugins",
    "ext_releaseref",
]
if "no_intersphinx" in tags:  # type: ignore[name-defined, ty:unresolved-reference]  # noqa: F821
    extensions.remove("sphinx.ext.intersphinx")

exclude_patterns = ["_build", "_applications.rst"]

templates_path = ["_templates"]


# -- Options for autodoc -------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html

autodoc_default_options = {
    "show-inheritance": True,
    "members": True,
    "member-order": "groupwise",  # autodoc_member_order
    "class-doc-from": "both",  # autoclass_content
}
autodoc_inherit_docstrings = False
autodoc_typehints = "description"


# -- Options for autosectionlabel ----------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autosectionlabel.html

autosectionlabel_prefix_document = True


# -- Options for intersphinx ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html

intersphinx_mapping = {
    # "python": ("https://docs.python.org/3", None),
    "requests": ("https://requests.readthedocs.io/en/stable/", None),
}

intersphinx_timeout = 60


# -- Options for ext_github ----------------------------------------------------
# file://./sphinxext/ext_github.py

github_project = "streamlink/streamlink"


# -- Options for ext_html_template_vars ----------------------------------------
# file://./sphinxext/ext_html_template_vars.py

html_template_vars = {
    "oneliner": (
        "A command-line utility that extracts streams from various services and pipes them into a video player of choice."
    ),
}


# -- Options for HTML output ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_theme_options = {
    "source_repository": "https://github.com/streamlink/streamlink/",
    "source_branch": "master",
    "source_directory": "docs/",
    "light_logo": "icon.svg",
    "dark_logo": "icon.svg",
}

html_logo = "../icon.svg"

html_css_files = [
    "styles/custom.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.3.0/css/fontawesome.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.3.0/css/solid.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.3.0/css/brands.min.css",
]

html_static_path = ["_static"]

html_sidebars = {
    "**": [
        "sidebar/scroll-start.html",
        "sidebar/brand.html",
        "sidebar/search.html",
        "sidebar/navigation.html",
        "sidebar/github-buttons.html",
        "sidebar/scroll-end.html",
    ],
}
if "no_github_buttons" in tags:  # type: ignore[name-defined, ty:unresolved-reference]  # noqa: F821
    html_sidebars.get("**", []).remove("sidebar/github-buttons.html")

html_domain_indices = False
html_show_sourcelink = False


# -- Options for manual page output --------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-manual-page-output

# Only include the man page in builds with the "man" tag set: via `-t man` (see Makefile)
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-tags
if "man" not in tags:  # type: ignore[name-defined, ty:unresolved-reference]  # noqa: F821
    exclude_patterns.append("_man.rst")

man_pages = [
    (
        "_man",
        "streamlink",
        "extracts streams from various services and pipes them into a video player of choice",
        ["Streamlink Contributors"],
        1,
    ),
]

man_make_section_directory = False



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import ANY, Mock, call

import pytest

import streamlink_cli.main
import tests


if TYPE_CHECKING:
    from streamlink import Streamlink


@pytest.fixture(autouse=True)
def session(session: Streamlink):
    session.plugins.load_path(Path(tests.__path__[0]) / "plugin")

    return session


@pytest.fixture(autouse=True)
def _player_setup(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("streamlink_cli.output.player.which", Mock(side_effect=lambda path: path))
    monkeypatch.setattr("streamlink_cli.output.player.sleep", Mock())


@pytest.fixture(autouse=True)
def mock_subprocess(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
    param = getattr(request, "param", {})
    popen = param.get("popen", [])
    call_ = param.get("call", [])

    mock_call = Mock()
    monkeypatch.setattr("streamlink_cli.output.player.subprocess.call", mock_call)

    mock_popen = Mock()
    mock_popen.return_value = Mock(poll=Mock(side_effect=[None, 0]))
    monkeypatch.setattr("streamlink_cli.output.player.subprocess.Popen", mock_popen)

    yield

    if popen:
        assert mock_popen.call_args_list == [call(popen, env=ANY, bufsize=ANY, stdin=ANY, stdout=ANY, stderr=ANY)]
        assert mock_call.call_args_list == []
    else:
        assert mock_popen.call_args_list == []
        assert mock_call.call_args_list == [call(call_, env=ANY, stdout=ANY, stderr=ANY)]


@pytest.fixture(autouse=True)
def _test(argv: list, mock_subprocess: Mock):
    with pytest.raises(SystemExit) as exc_info:
        streamlink_cli.main.main()
    assert exc_info.value.code == 0


@pytest.mark.parametrize(
    ("argv", "mock_subprocess"),
    [
        pytest.param(
            ["--player=player", "http://test.se", "test"],
            {"popen": ["player", "-"]},
            id="player-stdin",
        ),
        pytest.param(
            ["-p", "player", "-a", "-v {playerinput}", "http://test.se", "test"],
            {"popen": ["player", "-v", "-"]},
            id="player-args-single-hyphen-ghissue-971",
        ),
        pytest.param(
            ["-p", "player", "-a", '--input-title-format "foo \\"bar\\""', "--player-passthrough=hls", "test.se", "hls"],
            {"call": ["player", "--input-title-format", 'foo "bar"', "http://test.se/playlist.m3u8"]},
            id="player-passthrough",
        ),
    ],
    indirect=["argv", "mock_subprocess"],
)
def test_player_argv(argv: list, mock_subprocess: Mock):
    pass


@pytest.mark.posix_only()
@pytest.mark.parametrize(
    ("argv", "mock_subprocess"),
    [
        pytest.param(
            ["-p", "/usr/bin/vlc", "http://test.se", "test"],
            {"popen": ["/usr/bin/vlc", "--input-title-format", "http://test.se", "-"]},
            id="vlc-default",
        ),
        pytest.param(
            ["-p", "/usr/bin/vlc", "--title", "{title} - {author} - {category}", "http://test.se", "test"],
            {"popen": ["/usr/bin/vlc", "--input-title-format", "Test Title - Tѥst Āuƭhǿr - No Category", "-"]},
            id="vlc-custom",
        ),
        pytest.param(
            ["-p", "/Applications/VLC/vlc", "-a", "--play-and-exit", "http://test.se", "test"],
            {"popen": ["/Applications/VLC/vlc", "--input-title-format", "http://test.se", "--play-and-exit", "-"]},
            id="vlc-custom-with-args",
        ),
        pytest.param(
            ["-p", "/usr/bin/mpv", "--title", "{title}", "http://test.se", "test"],
            {"popen": ["/usr/bin/mpv", "--force-media-title=Test Title", "-"]},
            id="mpv-default",
        ),
    ],
    indirect=["argv", "mock_subprocess"],
)
def test_player_title_posix(argv: list, mock_subprocess: Mock):
    pass


@pytest.mark.windows_only()
@pytest.mark.parametrize(
    ("argv", "mock_subprocess"),
    [
        pytest.param(
            ["-p", "Z:\\VideoLAN\\vlc.exe", "http://test.se", "test"],
            {"popen": ["Z:\\VideoLAN\\vlc.exe", "--input-title-format", "http://test.se", "-"]},
            id="vlc-default",
        ),
        pytest.param(
            ["-p", "Z:\\VideoLAN\\vlc.exe", "--title", "{title} - {author} - {category}", "http://test.se", "test"],
            {"popen": ["Z:\\VideoLAN\\vlc.exe", "--input-title-format", "Test Title - Tѥst Āuƭhǿr - No Category", "-"]},
            id="vlc-custom",
        ),
        pytest.param(
            ["-p", "Z:\\VideoLAN\\vlc.exe", "-a", "--play-and-exit", "http://test.se", "test"],
            {"popen": ["Z:\\VideoLAN\\vlc.exe", "--input-title-format", "http://test.se", "--play-and-exit", "-"]},
            id="vlc-custom-with-args",
        ),
        pytest.param(
            ["-p", "Z:\\PotPlayerMini64.exe", "--player-passthrough=hls", "http://test.se/stream", "hls"],
            {"call": ["Z:\\PotPlayerMini64.exe", "http://test.se/playlist.m3u8\\http://test.se/stream"]},
            id="potplayer-passthrough-default",
        ),
        pytest.param(
            ["-p", "Z:\\PotPlayerMini64.exe", "--player-passthrough=hls", "--title", "{title}", "http://test.se/stream", "hls"],
            {"call": ["Z:\\PotPlayerMini64.exe", "http://test.se/playlist.m3u8\\Test Title"]},
            id="potplayer-passthrough-custom",
        ),
    ],
    indirect=["argv", "mock_subprocess"],
)
def test_player_title_windows(argv: list, mock_subprocess: Mock):
    pass



from __future__ import annotations

import re
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest

import tests.plugin
from streamlink.exceptions import NoPluginError, StreamlinkDeprecationWarning
from streamlink.options import Options
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


if TYPE_CHECKING:
    import requests_mock as rm

    from streamlink.session import Streamlink


PATH_TESTPLUGINS = Path(tests.plugin.__path__[0])


class TestPluginsDeprecations:
    def test_get_plugins(self, monkeypatch: pytest.MonkeyPatch, recwarn: pytest.WarningsRecorder, session: Streamlink):
        mock = Mock(return_value={})
        monkeypatch.setattr(session.plugins, "get_loaded", mock)
        assert session.get_plugins() == {}
        assert mock.call_count == 1
        assert [(record.filename, record.category, str(record.message)) for record in recwarn.list] == [
            (
                __file__,
                StreamlinkDeprecationWarning,
                "`Streamlink.get_plugins()` has been deprecated in favor of `Streamlink.plugins.get_loaded()`",
            ),
        ]

    def test_load_builtin_plugins(self, monkeypatch: pytest.MonkeyPatch, recwarn: pytest.WarningsRecorder, session: Streamlink):
        mock = Mock(return_value={})
        monkeypatch.setattr(session.plugins, "load_builtin", mock)
        assert session.load_builtin_plugins() is None
        assert mock.call_count == 1
        assert [(record.filename, record.category, str(record.message)) for record in recwarn.list] == [
            (
                __file__,
                StreamlinkDeprecationWarning,
                "`Streamlink.load_builtin_plugins()` has been deprecated in favor of the `plugins_builtin` keyword argument",
            ),
        ]

    def test_load_plugins(self, recwarn: pytest.WarningsRecorder, session: Streamlink):
        session.load_plugins(str(PATH_TESTPLUGINS))
        assert session.plugins.get_names() == ["testplugin"]
        assert session.plugins["testplugin"].__name__ == "TestPlugin"
        assert session.plugins["testplugin"].__module__ == "streamlink.plugins.testplugin"
        assert [(record.filename, record.category, str(record.message)) for record in recwarn.list] == [
            (
                __file__,
                StreamlinkDeprecationWarning,
                "`Streamlink.load_plugins()` has been deprecated in favor of `Streamlink.plugins.load_path()`",
            ),
        ]


class _EmptyPlugin(Plugin):
    def _get_streams(self):
        pass  # pragma: no cover


class TestResolveURL:
    @pytest.fixture(autouse=True)
    def _load_plugins(self, session: Streamlink):
        session.plugins.load_path(PATH_TESTPLUGINS)

    @pytest.fixture(autouse=True)
    def requests_mock(self, requests_mock: rm.Mocker):
        return requests_mock

    def test_resolve_url(self, recwarn: pytest.WarningsRecorder, session: Streamlink):
        _pluginname, pluginclass, resolved_url = session.resolve_url("http://test.se/channel")

        assert issubclass(pluginclass, Plugin)
        assert pluginclass is session.plugins["testplugin"]
        assert resolved_url == "http://test.se/channel"
        assert hasattr(session.resolve_url, "cache_info"), "resolve_url has a lookup cache"
        assert recwarn.list == []

    def test_resolve_url_noplugin(self, requests_mock: rm.Mocker, session: Streamlink):
        requests_mock.get("http://invalid2", status_code=301, headers={"Location": "http://invalid3"})

        with pytest.raises(NoPluginError):
            session.resolve_url("http://invalid1")
        with pytest.raises(NoPluginError):
            session.resolve_url("http://invalid2")

    def test_resolve_url_redirected(self, requests_mock: rm.Mocker, session: Streamlink):
        requests_mock.request("HEAD", "http://redirect1", status_code=501)
        requests_mock.request("GET", "http://redirect1", status_code=301, headers={"Location": "http://redirect2"})
        requests_mock.request("GET", "http://redirect2", status_code=301, headers={"Location": "http://test.se/channel"})
        requests_mock.request("GET", "http://test.se/channel", content=b"")

        _pluginname, pluginclass, resolved_url = session.resolve_url("http://redirect1")
        assert issubclass(pluginclass, Plugin)
        assert pluginclass is session.plugins["testplugin"]
        assert resolved_url == "http://test.se/channel"

    def test_resolve_url_no_redirect(self, session: Streamlink):
        _pluginname, pluginclass, resolved_url = session.resolve_url_no_redirect("http://test.se/channel")
        assert issubclass(pluginclass, Plugin)
        assert pluginclass is session.plugins["testplugin"]
        assert resolved_url == "http://test.se/channel"

    def test_resolve_url_no_redirect_noplugin(self, session: Streamlink):
        with pytest.raises(NoPluginError):
            session.resolve_url_no_redirect("http://invalid")

    def test_resolve_url_scheme(self, session: Streamlink):
        @pluginmatcher(re.compile(r"http://insecure"))
        class PluginHttp(_EmptyPlugin):
            pass

        @pluginmatcher(re.compile(r"https://secure"))
        class PluginHttps(_EmptyPlugin):
            pass

        session.plugins.update({
            "insecure": PluginHttp,
            "secure": PluginHttps,
        })

        with pytest.raises(NoPluginError):
            session.resolve_url("insecure")
        assert session.resolve_url("http://insecure")[1] is PluginHttp
        with pytest.raises(NoPluginError):
            session.resolve_url("https://insecure")

        assert session.resolve_url("secure")[1] is PluginHttps
        with pytest.raises(NoPluginError):
            session.resolve_url("http://secure")
        assert session.resolve_url("https://secure")[1] is PluginHttps


class TestStreams:
    @pytest.fixture(autouse=True)
    def _load_plugins(self, session: Streamlink):
        session.plugins.load_path(PATH_TESTPLUGINS)

    def test_streams(self, session: Streamlink):
        streams = session.streams("http://test.se/channel")

        assert "best" in streams
        assert "worst" in streams
        assert streams["best"] is streams["1080p"]
        assert streams["worst"] is streams["350k"]
        assert isinstance(streams["http"], HTTPStream)
        assert isinstance(streams["hls"], HLSStream)

    def test_streams_options(self, session: Streamlink):
        streams = session.streams("http://test.se/fromoptions", Options({"streamurl": "http://foo/"}))

        assert sorted(streams.keys()) == ["best", "fromoptions", "worst"]
        assert isinstance(streams["fromoptions"], HTTPStream)
        assert streams["fromoptions"].url == "http://foo/"

    def test_stream_types(self, session: Streamlink):
        streams = session.streams("http://test.se/channel", stream_types=["http", "hls"])
        assert isinstance(streams["480p"], HTTPStream)
        assert isinstance(streams["480p_hls"], HLSStream)

        streams = session.streams("http://test.se/channel", stream_types=["hls", "http"])
        assert isinstance(streams["480p"], HLSStream)
        assert isinstance(streams["480p_http"], HTTPStream)

    def test_stream_sorting_excludes(self, session: Streamlink):
        streams = session.streams("http://test.se/channel", sorting_excludes=[])
        assert "best" in streams
        assert "worst" in streams
        assert "best-unfiltered" not in streams
        assert "worst-unfiltered" not in streams
        assert streams["worst"] is streams["350k"]
        assert streams["best"] is streams["1080p"]

        streams = session.streams("http://test.se/channel", sorting_excludes=["1080p", "3000k"])
        assert "best" in streams
        assert "worst" in streams
        assert "best-unfiltered" not in streams
        assert "worst-unfiltered" not in streams
        assert streams["worst"] is streams["350k"]
        assert streams["best"] is streams["1500k"]

        streams = session.streams("http://test.se/channel", sorting_excludes=[">=1080p", ">1500k"])
        assert streams["best"] is streams["1500k"]

        streams = session.streams("http://test.se/channel", sorting_excludes=lambda q: not q.endswith("p"))
        assert streams["best"] is streams["3000k"]

        streams = session.streams("http://test.se/channel", sorting_excludes=lambda q: False)
        assert "best" not in streams
        assert "worst" not in streams
        assert "best-unfiltered" in streams
        assert "worst-unfiltered" in streams
        assert streams["worst-unfiltered"] is streams["350k"]
        assert streams["best-unfiltered"] is streams["1080p"]

        streams = session.streams("http://test.se/UnsortableStreamNames")
        assert "best" not in streams
        assert "worst" not in streams
        assert "best-unfiltered" not in streams
        assert "worst-unfiltered" not in streams
        assert "vod" in streams
        assert "vod_alt" in streams
        assert "vod_alt2" in streams



from __future__ import annotations

import contextlib
from contextlib import nullcontext
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING
from unittest.mock import AsyncMock

import pytest
import trio
from trio.testing import wait_all_tasks_blocked
from trio_websocket import CloseReason, ConnectionClosed, ConnectionTimeout

from streamlink.compat import ExceptionGroup
from streamlink.webbrowser.cdp.connection import MAX_BUFFER_SIZE, CDPConnection, CDPSession
from streamlink.webbrowser.cdp.devtools.target import SessionID, TargetID
from streamlink.webbrowser.cdp.devtools.util import CDPEvent
from streamlink.webbrowser.cdp.exceptions import CDPError
from tests.webbrowser.cdp import FakeWebsocketConnection


if TYPE_CHECKING:
    from collections.abc import Generator

    from trio.testing import MockClock
    from typing_extensions import Self

    from streamlink.webbrowser.cdp.connection import CDPEventListener
    from streamlink.webbrowser.cdp.devtools.util import T_JSON_DICT


EPSILON = 0.1


@dataclass
class FakeCommand(str):
    value: str

    def to_json(self) -> T_JSON_DICT:
        return {"value": self.value}

    @classmethod
    def from_json(cls, data: T_JSON_DICT):
        return cls(data["value"])


def fake_command(command: FakeCommand) -> Generator[T_JSON_DICT, T_JSON_DICT, FakeCommand]:
    json: T_JSON_DICT
    json = yield {"method": "Fake.fakeCommand", "params": command.to_json()}
    return FakeCommand.from_json(json)


def bad_command() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    yield {"method": "Fake.badCommand", "params": {}}
    yield {}


# The event name registration is monkey-patched by the test fixture, so we leave it empty
@dataclass
class FakeEvent(CDPEvent, event=None):
    value: str

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Self:
        return cls(json["value"])


@pytest.fixture()
async def cdp_connection(websocket_connection: FakeWebsocketConnection):
    try:
        async with CDPConnection.create("ws://localhost:1234/fake") as cdp_connection:
            assert isinstance(cdp_connection, CDPConnection)
            assert not websocket_connection.closed
            try:
                yield cdp_connection
            finally:
                await cdp_connection.aclose()
                assert cdp_connection.sessions == {}
    finally:
        assert websocket_connection.closed


class TestCreateConnection:
    @pytest.mark.trio()
    async def test_success(self, cdp_connection: CDPConnection):
        assert cdp_connection.target_id is None
        assert cdp_connection.session_id is None

    @pytest.mark.trio()
    async def test_failure(self, monkeypatch: pytest.MonkeyPatch):
        fake_connect_websocket_url = AsyncMock(side_effect=ConnectionTimeout)
        monkeypatch.setattr("streamlink.webbrowser.cdp.connection.connect_websocket_url", fake_connect_websocket_url)
        with pytest.raises(ExceptionGroup) as excinfo:
            async with CDPConnection.create("ws://localhost:1234/fake"):
                pass  # pragma: no cover
        assert excinfo.group_contains(ConnectionTimeout)

    @pytest.mark.trio()
    @pytest.mark.parametrize(
        ("timeout", "expected"),
        [
            pytest.param(None, 2, id="Default value of 2 seconds"),
            pytest.param(0, 2, id="No timeout uses default value"),
            pytest.param(3, 3, id="Custom timeout value"),
        ],
    )
    async def test_timeout(self, websocket_connection: FakeWebsocketConnection, timeout: int | None, expected: int):
        async with CDPConnection.create("ws://localhost:1234/fake", timeout=timeout) as cdp_conn:
            pass
        assert cdp_conn.cmd_timeout == expected


class TestReaderError:
    @pytest.mark.trio()
    async def test_invalid_json(self, caplog: pytest.LogCaptureFixture, websocket_connection: FakeWebsocketConnection):
        with pytest.raises(ExceptionGroup) as excinfo:  # noqa: PT012
            async with CDPConnection.create("ws://localhost:1234/fake"):
                assert not websocket_connection.closed
                await websocket_connection.sender.send("INVALID JSON")
                await wait_all_tasks_blocked()

        assert excinfo.group_contains(
            CDPError,
            match=r"^Received invalid CDP JSON data: Expecting value: line 1 column 1 \(char 0\)$",
        )
        assert caplog.records == []

    @pytest.mark.trio()
    async def test_unknown_session_id(self, caplog: pytest.LogCaptureFixture, websocket_connection: FakeWebsocketConnection):
        with pytest.raises(ExceptionGroup) as excinfo:  # noqa: PT012
            async with CDPConnection.create("ws://localhost:1234/fake"):
                assert not websocket_connection.closed
                await websocket_connection.sender.send("""{"sessionId":"unknown"}""")
                await wait_all_tasks_blocked()

        assert excinfo.group_contains(CDPError, match=r"^Unknown CDP session ID: SessionID\('unknown'\)$")
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            ("streamlink.webbrowser.cdp.connection", "all", """Received message: {"sessionId":"unknown"}"""),
        ]


@contextlib.contextmanager
def raises_group(*group_contains):
    try:
        with pytest.raises(ExceptionGroup) as excinfo:
            yield
    finally:
        for args, kwargs, expected in group_contains:
            assert excinfo.group_contains(*args, **kwargs) is expected


class TestSend:
    # noinspection PyUnusedLocal
    @pytest.mark.trio()
    @pytest.mark.parametrize(
        ("timeout", "jump", "raises"),
        [
            pytest.param(
                None,
                2 - EPSILON,
                nullcontext(),
                id="Default timeout, response in time",
            ),
            pytest.param(
                None,
                2,
                raises_group(
                    ((CDPError,), {"match": "^Sending CDP message and receiving its response timed out$"}, True),
                ),
                id="Default timeout, response not in time",
            ),
            pytest.param(
                3,
                3 - EPSILON,
                nullcontext(),
                id="Custom timeout, response in time",
            ),
            pytest.param(
                3,
                3,
                raises_group(
                    ((CDPError,), {"match": "^Sending CDP message and receiving its response timed out$"}, True),
                ),
                id="Custom timeout, response not in time",
            ),
        ],
    )
    async def test_timeout(
        self,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
        autojump_clock: MockClock,
        timeout: float | None,
        jump: float,
        raises: nullcontext,
    ):
        assert cdp_connection.cmd_timeout == 2
        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == []

        async def response():
            await trio.sleep(jump)
            await websocket_connection.sender.send("""{"id":0,"result":{"value":"foo"}}""")

        with raises:
            async with trio.open_nursery() as nursery:
                nursery.start_soon(partial(cdp_connection.send, fake_command(FakeCommand("foo")), timeout=timeout))
                nursery.start_soon(response)

        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == ["""{"id":0,"method":"Fake.fakeCommand","params":{"value":"foo"}}"""]

    @pytest.mark.trio()
    async def test_closed(
        self,
        monkeypatch: pytest.MonkeyPatch,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
    ):
        # noinspection PyTypeChecker
        fake_send_message = AsyncMock(side_effect=ConnectionClosed(CloseReason(1000, None)))
        monkeypatch.setattr(FakeWebsocketConnection, "send_message", fake_send_message)

        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == []

        with pytest.raises(CDPError) as cm:
            await cdp_connection.send(fake_command(FakeCommand("foo")))

        assert str(cm.value) == "CloseReason<code=1000, name=NORMAL_CLOSURE, reason=None>"
        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == []
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Sending message: {"id":0,"method":"Fake.fakeCommand","params":{"value":"foo"}}""",
            ),
        ]

    @pytest.mark.trio()
    async def test_bad_command(
        self,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
    ):
        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == []

        with pytest.raises(ExceptionGroup) as excinfo:  # noqa: PT012
            async with trio.open_nursery() as nursery:
                nursery.start_soon(cdp_connection.send, bad_command())
                nursery.start_soon(websocket_connection.sender.send, """{"id":0,"result":{}}""")

        assert excinfo.group_contains(CDPError, match="^Generator of CDP command ID 0 did not exit when expected!$")
        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == ["""{"id":0,"method":"Fake.badCommand","params":{}}"""]
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Sending message: {"id":0,"method":"Fake.badCommand","params":{}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"id":0,"result":{}}""",
            ),
        ]

    @pytest.mark.trio()
    async def test_result_exception(
        self,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
    ):
        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == []

        with pytest.raises(ExceptionGroup) as excinfo:  # noqa: PT012
            async with trio.open_nursery() as nursery:
                nursery.start_soon(cdp_connection.send, fake_command(FakeCommand("foo")))
                nursery.start_soon(websocket_connection.sender.send, """{"id":0,"result":{}}""")

        assert excinfo.group_contains(CDPError, match="^Generator of CDP command ID 0 raised KeyError: 'value'$")
        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == ["""{"id":0,"method":"Fake.fakeCommand","params":{"value":"foo"}}"""]
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Sending message: {"id":0,"method":"Fake.fakeCommand","params":{"value":"foo"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"id":0,"result":{}}""",
            ),
        ]

    @pytest.mark.trio()
    async def test_result_success(
        self,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
    ):
        results = {}

        async def send(key):
            results[key] = await cdp_connection.send(fake_command(FakeCommand(key)))

        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == []

        async with trio.open_nursery() as nursery:
            # ensure that we start the tasks in the correct order
            nursery.start_soon(send, "foo")
            await wait_all_tasks_blocked()
            nursery.start_soon(send, "bar")
            await wait_all_tasks_blocked()

            assert list(cdp_connection.cmd_buffers.keys()) == [0, 1]
            assert all(buf.response is None for buf in cdp_connection.cmd_buffers.values())
            assert all(buf.event.is_set() is False for buf in cdp_connection.cmd_buffers.values())

            # send result of second command first
            nursery.start_soon(websocket_connection.sender.send, """{"id":1,"result":{"value":"BAR"}}""")
            await wait_all_tasks_blocked()
            nursery.start_soon(websocket_connection.sender.send, """{"id":0,"result":{"value":"FOO"}}""")

        assert list(results.keys()) == ["bar", "foo"]
        assert all(isinstance(result, FakeCommand) for result in results.values())
        assert results["foo"].value == "FOO"
        assert results["bar"].value == "BAR"

        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == [
            """{"id":0,"method":"Fake.fakeCommand","params":{"value":"foo"}}""",
            """{"id":1,"method":"Fake.fakeCommand","params":{"value":"bar"}}""",
        ]
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Sending message: {"id":0,"method":"Fake.fakeCommand","params":{"value":"foo"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Sending message: {"id":1,"method":"Fake.fakeCommand","params":{"value":"bar"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"id":1,"result":{"value":"BAR"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"id":0,"result":{"value":"FOO"}}""",
            ),
        ]


class TestHandleCmdResponse:
    @pytest.mark.trio()
    async def test_unknown_id(
        self,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
    ):
        assert cdp_connection.cmd_buffers == {}

        await websocket_connection.sender.send("""{"id":123}""")
        await wait_all_tasks_blocked()

        assert cdp_connection.cmd_buffers == {}
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"id":123}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "warning",
                "Got a CDP command response with an unknown ID: 123",
            ),
        ]

    @pytest.mark.trio()
    async def test_response_error(
        self,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
    ):
        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == []

        with pytest.raises(ExceptionGroup) as excinfo:  # noqa: PT012
            async with trio.open_nursery() as nursery:
                nursery.start_soon(cdp_connection.send, fake_command(FakeCommand("foo")))
                nursery.start_soon(websocket_connection.sender.send, """{"id":0,"error":"Some error message"}""")

        assert excinfo.group_contains(CDPError, match="^Error in CDP command response 0: Some error message$")
        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == ["""{"id":0,"method":"Fake.fakeCommand","params":{"value":"foo"}}"""]
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Sending message: {"id":0,"method":"Fake.fakeCommand","params":{"value":"foo"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"id":0,"error":"Some error message"}""",
            ),
        ]

    @pytest.mark.trio()
    async def test_response_no_result(
        self,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
    ):
        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == []

        with pytest.raises(ExceptionGroup) as excinfo:  # noqa: PT012
            async with trio.open_nursery() as nursery:
                nursery.start_soon(cdp_connection.send, fake_command(FakeCommand("foo")))
                nursery.start_soon(websocket_connection.sender.send, """{"id":0}""")

        assert excinfo.group_contains(CDPError, match="^No result in CDP command response 0$")
        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == ["""{"id":0,"method":"Fake.fakeCommand","params":{"value":"foo"}}"""]
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Sending message: {"id":0,"method":"Fake.fakeCommand","params":{"value":"foo"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"id":0}""",
            ),
        ]


class TestSession:
    @pytest.fixture()
    async def cdp_session(self, cdp_connection: CDPConnection):
        target_id = TargetID("01234")
        session_id = SessionID("56789")
        session = cdp_connection.sessions[session_id] = CDPSession(
            cdp_connection.websocket,
            target_id=target_id,
            session_id=session_id,
            cmd_timeout=cdp_connection.cmd_timeout,
        )
        return session

    @pytest.mark.trio()
    async def test_new_target(
        self,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
    ):
        assert cdp_connection.sessions == {}
        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == []

        session = None

        async def send():
            nonlocal session
            session = await cdp_connection.new_target("http://localhost")

        async with trio.open_nursery() as nursery:
            nursery.start_soon(send)
            await wait_all_tasks_blocked()
            nursery.start_soon(websocket_connection.sender.send, """{"id":0,"result":{"targetId":"01234"}}""")
            await wait_all_tasks_blocked()
            nursery.start_soon(websocket_connection.sender.send, """{"id":1,"result":{"sessionId":"56789"}}""")

        assert isinstance(session, CDPSession)
        assert session.target_id == TargetID("01234")
        assert session.session_id == SessionID("56789")
        assert cdp_connection.sessions[SessionID("56789")] is session

        assert cdp_connection.cmd_buffers == {}
        assert websocket_connection.sent == [
            """{"id":0,"method":"Target.createTarget","params":{"url":"http://localhost"}}""",
            """{"id":1,"method":"Target.attachToTarget","params":{"flatten":true,"targetId":"01234"}}""",
        ]
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Sending message: {"id":0,"method":"Target.createTarget","params":{"url":"http://localhost"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"id":0,"result":{"targetId":"01234"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Sending message: {"id":1,"method":"Target.attachToTarget","params":{"flatten":true,"targetId":"01234"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"id":1,"result":{"sessionId":"56789"}}""",
            ),
        ]

    @pytest.mark.trio()
    async def test_session_command(
        self,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        cdp_session: CDPSession,
        websocket_connection: FakeWebsocketConnection,
    ):
        results = {}

        async def send(obj, key):
            results[key] = await obj.send(fake_command(FakeCommand(key)))

        assert cdp_connection.cmd_buffers == {}
        assert cdp_session.cmd_buffers == {}
        assert websocket_connection.sent == []

        async with trio.open_nursery() as nursery:
            # ensure that we start the tasks in the correct order
            nursery.start_soon(send, cdp_connection, "foo")
            await wait_all_tasks_blocked()
            nursery.start_soon(send, cdp_session, "bar")
            await wait_all_tasks_blocked()

            assert list(cdp_connection.cmd_buffers.keys()) == [0]
            assert list(cdp_session.cmd_buffers.keys()) == [0]

            # send result of second command first
            nursery.start_soon(websocket_connection.sender.send, """{"id":0,"result":{"value":"BAR"},"sessionId":"56789"}""")
            await wait_all_tasks_blocked()
            nursery.start_soon(websocket_connection.sender.send, """{"id":0,"result":{"value":"FOO"}}""")

        assert list(results.keys()) == ["bar", "foo"]
        assert all(isinstance(result, FakeCommand) for result in results.values())
        assert results["foo"].value == "FOO"
        assert results["bar"].value == "BAR"

        assert cdp_connection.cmd_buffers == {}
        assert cdp_session.cmd_buffers == {}
        assert websocket_connection.sent == [
            """{"id":0,"method":"Fake.fakeCommand","params":{"value":"foo"}}""",
            """{"id":0,"method":"Fake.fakeCommand","params":{"value":"bar"},"sessionId":"56789"}""",
        ]
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Sending message: {"id":0,"method":"Fake.fakeCommand","params":{"value":"foo"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Sending message: {"id":0,"method":"Fake.fakeCommand","params":{"value":"bar"},"sessionId":"56789"}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"id":0,"result":{"value":"BAR"},"sessionId":"56789"}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"id":0,"result":{"value":"FOO"}}""",
            ),
        ]


class TestHandleEvent:
    @pytest.fixture(autouse=True)
    def _event_parsers(self, monkeypatch: pytest.MonkeyPatch):
        event_parsers: dict[str, type] = {
            "Fake.fakeEvent": FakeEvent,
        }
        monkeypatch.setattr("streamlink.webbrowser.cdp.devtools.util._event_parsers", event_parsers)

    @pytest.mark.trio()
    @pytest.mark.parametrize(
        "message",
        [
            pytest.param("""{"foo":"bar"}""", id="Missing method and params"),
            pytest.param("""{"method":"method"}""", id="Missing params"),
            pytest.param("""{"params":{}}""", id="Missing method"),
        ],
    )
    async def test_invalid_event(
        self,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
        message: str,
    ):
        await websocket_connection.sender.send(message)
        await wait_all_tasks_blocked()
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                f"Received message: {message}",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "warning",
                "Invalid CDP event message received without method or params",
            ),
        ]

    @pytest.mark.trio()
    async def test_unknown_event(
        self,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
    ):
        await websocket_connection.sender.send("""{"method":"unknown","params":{}}""")
        await wait_all_tasks_blocked()
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"method":"unknown","params":{}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "warning",
                "Unknown CDP event message received: unknown",
            ),
        ]

    @pytest.mark.trio()
    async def test_eventlistener(
        self,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
    ):
        assert FakeEvent not in cdp_connection.event_channels
        listener1 = cdp_connection.listen(FakeEvent)
        listener2 = cdp_connection.listen(FakeEvent)
        listener3 = cdp_connection.listen(FakeEvent, max_buffer_size=MAX_BUFFER_SIZE * 2)
        listeners = listener1, listener2, listener3
        assert FakeEvent in cdp_connection.event_channels
        assert len(cdp_connection.event_channels[FakeEvent]) == 3

        assert listener1._sender.statistics().max_buffer_size == MAX_BUFFER_SIZE
        assert listener2._sender.statistics().max_buffer_size == MAX_BUFFER_SIZE
        assert listener3._sender.statistics().max_buffer_size == MAX_BUFFER_SIZE * 2

        results = []

        async def listen_once(listener: CDPEventListener):
            async with listener as result:
                results.append(result)

        async def listen_twice(listener: CDPEventListener):
            async with listener as result:
                results.append(result)
                results.append(await listener.receive())

        async def listen_forever(listener: CDPEventListener):
            async for result in listener:
                results.append(result)

        async with trio.open_nursery() as nursery:
            nursery.start_soon(listen_once, listener1)
            await wait_all_tasks_blocked()
            nursery.start_soon(listen_twice, listener2)
            await wait_all_tasks_blocked()
            nursery.start_soon(listen_forever, listener3)
            await wait_all_tasks_blocked()

            await websocket_connection.sender.send("""{"method":"Fake.fakeEvent","params":{"value":"foo"}}""")
            await wait_all_tasks_blocked()
            assert len(cdp_connection.event_channels[FakeEvent]) == 3

            await websocket_connection.sender.send("""{"method":"Fake.fakeEvent","params":{"value":"bar"}}""")
            await wait_all_tasks_blocked()
            assert len(cdp_connection.event_channels[FakeEvent]) == 2

            await websocket_connection.sender.send("""{"method":"Fake.fakeEvent","params":{"value":"baz"}}""")
            await wait_all_tasks_blocked()
            assert len(cdp_connection.event_channels[FakeEvent]) == 1

            await cdp_connection.aclose()

        assert results == [
            FakeEvent(value="foo"),
            FakeEvent(value="foo"),
            FakeEvent(value="foo"),
            FakeEvent(value="bar"),
            FakeEvent(value="bar"),
            FakeEvent(value="baz"),
        ]
        assert FakeEvent not in cdp_connection.event_channels
        assert all(listener._sender._closed for listener in listeners)  # type: ignore[attr-defined, ty:unresolved-attribute]
        assert all(listener._receiver._closed for listener in listeners)  # type: ignore[attr-defined, ty:unresolved-attribute]
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"method":"Fake.fakeEvent","params":{"value":"foo"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                "Received event: FakeEvent(value='foo')",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"method":"Fake.fakeEvent","params":{"value":"bar"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                "Received event: FakeEvent(value='bar')",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"method":"Fake.fakeEvent","params":{"value":"baz"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                "Received event: FakeEvent(value='baz')",
            ),
        ]

    @pytest.mark.trio()
    async def test_would_block(
        self,
        caplog: pytest.LogCaptureFixture,
        cdp_connection: CDPConnection,
        websocket_connection: FakeWebsocketConnection,
    ):
        assert FakeEvent not in cdp_connection.event_channels
        listener = cdp_connection.listen(FakeEvent, max_buffer_size=1)
        assert FakeEvent in cdp_connection.event_channels
        assert len(cdp_connection.event_channels[FakeEvent]) == 1

        assert listener._sender.statistics().current_buffer_used == 0
        assert listener._sender.statistics().max_buffer_size == 1

        await websocket_connection.sender.send("""{"method":"Fake.fakeEvent","params":{"value":"foo"}}""")
        await wait_all_tasks_blocked()
        assert listener._sender.statistics().current_buffer_used == 1
        assert listener._sender.statistics().max_buffer_size == 1

        await websocket_connection.sender.send("""{"method":"Fake.fakeEvent","params":{"value":"bar"}}""")
        await wait_all_tasks_blocked()
        assert listener._sender.statistics().current_buffer_used == 1
        assert listener._sender.statistics().max_buffer_size == 1

        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"method":"Fake.fakeEvent","params":{"value":"foo"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                "Received event: FakeEvent(value='foo')",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                """Received message: {"method":"Fake.fakeEvent","params":{"value":"bar"}}""",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "all",
                "Received event: FakeEvent(value='bar')",
            ),
            (
                "streamlink.webbrowser.cdp.connection",
                "error",
                """Unable to propagate CDP event FakeEvent(value='bar') due to full channel""",
            ),
        ]



from streamlink.plugins.mediaklikk import Mediaklikk
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlMediaklikk(PluginCanHandleUrl):
    __plugin__ = Mediaklikk

    should_match = [
        "https://www.mediaklikk.hu/duna-world-elo/",
        "https://www.mediaklikk.hu/m1-elo",
        "https://www.mediaklikk.hu/m2-elo",
        "https://m4sport.hu/elo/",
        "https://m4sport.hu/elo/?channelId=m4sport+",
        "https://m4sport.hu/elo/?showchannel=mtv4plus",
        "https://m4sport.hu/video/2025/08/08/fizz-liga-kolorcity-kazincbarcika-sc-dvsc-merkozes",
        "https://hirado.hu/elo/m1",
        "https://hirado.hu/elo/m2",
        "https://hirado.hu/belfold/video/2025/08/09/szuletesnap-a-budapesti-allat-es-novenykertben",
    ]



from streamlink.plugins.dailymotion import DailyMotion
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlDailyMotion(PluginCanHandleUrl):
    __plugin__ = DailyMotion

    should_match_groups = [
        (("user", "https://www.dailymotion.com/france24"), {"user": "france24"}),
        (("media", "https://www.dailymotion.com/video/x8dmdzz"), {"media_id": "x8dmdzz"}),
        (("media", "https://www.dailymotion.com/embed/video/x8dmdzz"), {"media_id": "x8dmdzz"}),
        (("lequipe", "https://www.lequipe.fr/tv/videos/live/k3HiS3JB0BsORKqwC49"), {}),
        (("lequipe", "https://www.lequipe.fr/tv/replay/le-resume-de-la-mass-start-du-grand-bornand/20201780"), {}),
    ]

    should_not_match = [
        "https://www.dailymotion.com/",
    ]



from __future__ import annotations

import threading
from typing import TYPE_CHECKING
from unittest.mock import Mock, call

import pytest

from streamlink.utils.named_pipe import NamedPipe, NamedPipePosix, NamedPipeWindows


try:
    from ctypes import byref, c_ulong, create_string_buffer, windll  # type: ignore[attr-defined]
except ImportError:
    pass


if TYPE_CHECKING:
    from streamlink.utils.named_pipe import NamedPipeBase


GENERIC_READ = 0x80000000
OPEN_EXISTING = 3


class ReadNamedPipeThread(threading.Thread):
    def __init__(self, pipe: NamedPipeBase):
        super().__init__(daemon=True)
        self.path = pipe.path
        self.error = None
        self.data = b""
        self.done = threading.Event()

    def run(self):
        try:
            self.read()
        except OSError as err:  # pragma: no cover
            self.error = err
        self.done.set()

    def read(self):
        raise NotImplementedError


class ReadNamedPipeThreadPosix(ReadNamedPipeThread):
    def read(self):
        with self.path.open("rb") as file:
            while True:
                data = file.read(-1)
                if len(data) == 0:
                    break
                self.data += data


class ReadNamedPipeThreadWindows(ReadNamedPipeThread):
    def read(self):
        handle = windll.kernel32.CreateFileW(str(self.path), GENERIC_READ, 0, None, OPEN_EXISTING, 0, None)
        try:
            while True:
                data = create_string_buffer(NamedPipeWindows.bufsize)
                read = c_ulong(0)
                if not windll.kernel32.ReadFile(handle, data, NamedPipeWindows.bufsize, byref(read), None):  # pragma: no cover
                    raise OSError(f"Failed reading pipe: {windll.kernel32.GetLastError()}")
                self.data += data.value
                if read.value != len(data.value):
                    break
        finally:
            windll.kernel32.CloseHandle(handle)


class TestNamedPipe:
    def test_name(self, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture):
        caplog.set_level(1, "streamlink")
        monkeypatch.setattr("streamlink.utils.named_pipe._id", 0)
        monkeypatch.setattr("streamlink.utils.named_pipe.os.getpid", Mock(return_value=12345))
        monkeypatch.setattr("streamlink.utils.named_pipe.random.randint", Mock(return_value=67890))
        monkeypatch.setattr("streamlink.utils.named_pipe.NamedPipe._create", Mock(return_value=None))

        NamedPipe()
        NamedPipe()

        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            ("streamlink.utils.named_pipe", "info", "Creating pipe streamlinkpipe-12345-1-67890"),
            ("streamlink.utils.named_pipe", "info", "Creating pipe streamlinkpipe-12345-2-67890"),
        ]


@pytest.mark.posix_only()
class TestNamedPipePosix:
    def test_export(self):
        assert NamedPipe is NamedPipePosix

    def test_create(self, monkeypatch: pytest.MonkeyPatch):
        mock_mkfifo = Mock(side_effect=OSError)
        monkeypatch.setattr("streamlink.utils.named_pipe.os.mkfifo", mock_mkfifo)
        with pytest.raises(OSError):  # noqa: PT011
            NamedPipePosix()
        assert mock_mkfifo.call_args[0][1:] == (0o660,)

    def test_close_before_open(self):
        pipe = NamedPipePosix()
        assert pipe.path.is_fifo()
        pipe.close()
        assert not pipe.path.is_fifo()
        # closing twice doesn't raise
        pipe.close()

    def test_close_error(self, monkeypatch: pytest.MonkeyPatch):
        mock_fd_close = Mock(side_effect=OSError)
        mock_fd = Mock(close=mock_fd_close)
        monkeypatch.setattr("pathlib.Path.open", Mock(return_value=mock_fd))

        pipe = NamedPipePosix()
        assert pipe.path.is_fifo()
        pipe.open()
        assert mock_fd_close.call_args_list == []

        with pytest.raises(OSError):  # noqa: PT011
            pipe.close()
        assert mock_fd_close.call_args_list == [call()]
        assert not pipe.path.is_fifo()

    def test_write_before_open(self):
        pipe = NamedPipePosix()
        assert pipe.path.is_fifo()
        with pytest.raises(AttributeError):
            pipe.write(b"foo")
        pipe.close()

    def test_named_pipe(self):
        pipe = NamedPipePosix()
        assert pipe.path.is_fifo()
        reader = ReadNamedPipeThreadPosix(pipe)
        reader.start()
        pipe.open()
        assert pipe.write(b"foo") == 3
        assert pipe.write(b"bar") == 3
        pipe.close()
        assert not pipe.path.is_fifo()
        reader.done.wait(4)
        reader.join(1)
        assert not reader.is_alive()
        assert reader.error is None
        assert reader.data == b"foobar"


@pytest.mark.windows_only()
class TestNamedPipeWindows:
    def test_export(self):
        assert NamedPipe is NamedPipeWindows

    def test_create(self, monkeypatch: pytest.MonkeyPatch):
        mock_kernel32 = Mock()
        mock_kernel32.CreateNamedPipeW.return_value = NamedPipeWindows.INVALID_HANDLE_VALUE
        mock_kernel32.GetLastError.return_value = 12345
        monkeypatch.setattr("streamlink.utils.named_pipe.windll.kernel32", mock_kernel32)

        with pytest.raises(OSError, match=r"^Named pipe error code 0x00003039$"):
            NamedPipeWindows()

        assert mock_kernel32.CreateNamedPipeW.call_args[0][1:] == (
            0x00000002,
            0x00000000,
            255,
            8192,
            8192,
            0,
            None,
        )

    def test_close_before_open(self):
        pipe = NamedPipeWindows()
        handle = windll.kernel32.CreateFileW(str(pipe.path), GENERIC_READ, 0, None, OPEN_EXISTING, 0, None)
        assert handle != NamedPipeWindows.INVALID_HANDLE_VALUE
        windll.kernel32.CloseHandle(handle)
        pipe.close()
        handle = windll.kernel32.CreateFileW(str(pipe.path), GENERIC_READ, 0, None, OPEN_EXISTING, 0, None)
        assert handle == NamedPipeWindows.INVALID_HANDLE_VALUE
        # closing twice doesn't raise
        pipe.close()

    @pytest.mark.parametrize("method", ["DisconnectNamedPipe", "CloseHandle"])
    def test_close_error(self, monkeypatch: pytest.MonkeyPatch, method: str):
        mock_method = Mock(side_effect=OSError)
        mock_kernel32 = Mock(**{method: mock_method})
        monkeypatch.setattr("streamlink.utils.named_pipe.windll.kernel32", mock_kernel32)

        pipe = NamedPipeWindows()
        mock_pipe = pipe.pipe
        assert mock_pipe is not None
        pipe.open()
        assert mock_method.call_args_list == []

        with pytest.raises(OSError):  # noqa: PT011
            pipe.close()
        assert mock_method.call_args_list == [call(mock_pipe)]
        assert pipe.pipe is None

    def test_named_pipe(self):
        pipe = NamedPipeWindows()
        reader = ReadNamedPipeThreadWindows(pipe)
        reader.start()
        pipe.open()
        assert pipe.write(b"foo") == 3
        assert pipe.write(b"bar") == 3
        assert pipe.write(b"\x00") == 1
        reader.done.wait(4)
        reader.join(1)
        assert not reader.is_alive()
        assert reader.error is None
        assert reader.data == b"foobar"
        pipe.close()
